(* AbsHeap.v: abstract heap states and properties thereupon. *)

Require Export SfLib.
Require Export ImpHeap.

(* This file implements the data types and operations on the abstract state,
   the \Pi | \Sigma of the paper. *)


(* Simple auxiliary list functions for lists of id's/nats *)

Fixpoint mem_id (e : id) (l : list id) { struct l } : bool :=
    match l with
        | (f::r) => if beq_id f e then true else mem_id e r
        | Nil    => false
    end.

Fixpoint mem_nat (e : nat) (l : list nat) { struct l } : bool :=
    match l with
        | (f::r) => if beq_nat f e then true else mem_nat e r
        | Nil    => false
    end.

Fixpoint merge (l1 : list id) (l2 : list id) { struct l1 } : list id :=
    match l1 with
       | a::b => if mem_id a l2 then merge b l2 else a::(merge b l2)
       | Nil => l2
    end.


(* ******************************************************************************
 *
 * absHeap, the abstract heaps, correspond to the \Sigma expressions from the paper.
 *
 *******************************************************************************)

Inductive absHeap : Type :=
    | Empty : absHeap
    | Compose : absHeap -> absHeap -> absHeap
    | HeapImply : absHeap -> absHeap -> absHeap
    | SpatialCell : aexp -> (list (id * aexp)) -> absHeap
    | R: list id -> aexp -> list aexp -> absHeap
    | Junk : absHeap.

Notation "x '**' y" := (Compose x y)
  (at level 30).

Notation "x '-*' y" := (HeapImply x y)
  (at level 40).

Notation "'[[' e m ']]'" := (SpatialCell e m)
  (at level 70).


(********************************************************************************
 *
 *  pi and formula are the \Pi and Body from the paper, respectively.
 *
 *******************************************************************************)

Inductive formula : Type :=
    | FTrue : formula
    | FFalse : formula
    | FEqual : aexp -> aexp -> formula
    | FLessEqual : aexp -> aexp -> formula
    | FAnd : formula -> formula -> formula
    | FOr : formula -> formula -> formula
    | FNot : formula -> formula
    | FAll : id -> absHeap -> formula -> formula
    | FExists : id -> absHeap -> formula -> formula
    | LeadsTo : aexp -> id -> aexp -> formula.

Inductive pi: Type :=
    | PBody : formula -> pi
    | PMember : id -> absHeap -> pi -> pi.

Notation "x '====' y" := (FEqual x y)
  (at level 40).

Notation "x '<<==' y" := (FLessEqual x y)
  (at level 40).

Notation "x '&&&&' y" := (FAnd x y)
  (at level 30).

Notation "x '|||||' y" := (FOr x y)
  (at level 20).

Notation "'fALL' v 'in' h ',' f" := (FAll v h f)
  (at level 25).

Notation "'fExists' v 'in' h ',' f" := (FExists v h f)
  (at level 25).

Notation "x '~' f '~>' y" := (LeadsTo x f y)
  (at level 50).

(********************************************************************************
 *
 * absState, an abstract state, is written as \Pi | \Sigma in the paper.
 *
 *******************************************************************************)

Definition absState := prod absHeap pi.

(********************************************************************************
 *
 * Here are some auxiliary definitions used in mapping abstract
 * to concrete states.
 *
 *******************************************************************************)

(* disjoint heaps *)

Definition disjoint (h1 : heap) (h2 : heap) : Prop :=
    forall (x : nat), h1 x=None \/ h2 x=None.

(* A binding is set of variable bindings, corresponding to the s \in Stack of the paper.
   Note that a binding is a partial version of an s, it may not map some id's so it 
   is not exactly the same as an s. *)

Definition binding := id -> (option nat).

Definition empty_binding : binding := (fun x => None).

Definition addBinding (b : binding) (v : id) (n : nat) : binding :=
    (fun x => if beq_id x v then Some n else b x).

(* disjoint bindings *)

Definition compatible (b1 : binding) (b2 : binding) : Prop :=
    forall (x : id), (b1 x=None) \/ (b2 x=None) \/ b1 x=b2 x.

Definition mergeHeaps (h1: heap) (h2: heap) : heap :=
    (fun x => match (h1 x) with
                  | None => h2 x
                  | Some x => Some x
              end).

Definition mergeBindings (b1: binding) (b2: binding) : binding :=
    (fun x => match (b1 x) with
                  | None => b2 x
                  | Some x => Some x
              end).

(* Lookup arith exp e in arith exp list l *)

Fixpoint mem_exp (e : aexp) (l : list aexp) { struct l } : bool :=
    match l with
        | (f::r) => if beq_exp f e then true else mem_exp e r
        | Nil    => false
    end.

(* A predicate to abstractly evaluate an arith exp.  Viewed 
   as a function it is partial - a variable may not have a binding.
   This is the \cal E definition in the paper. *)

Inductive absAExpEval : binding -> aexp -> nat -> Prop :=
    | ABENum : forall b n, absAExpEval b (ANum n) n
    | ABEVar : forall b v n,
               b v = Some n ->
               absAExpEval b (AVar v) n
    | ABEPlus : forall e1 e2 b n1 n2,
                absAExpEval b e1 n1 ->
                absAExpEval b e2 n2 ->
                absAExpEval b (APlus e1 e2) (plus n1 n2)
    | ABEMinus : forall e1 e2 b n1 n2,
                 absAExpEval b e1 n1 ->
                 absAExpEval b e2 n2 ->
                 absAExpEval b (AMinus e1 e2) (minus n1 n2)
    | ABETimes : forall e1 e2 b n1 n2,
                 absAExpEval b e1 n1 ->
                 absAExpEval b e2 n2 ->
                 absAExpEval b (AMult e1 e2) (mult n1 n2).

(* \cal E over a list of expressions *)

Inductive absAExpEvalList : binding -> list aexp -> list nat -> Prop :=
    | ABELNil : forall b, absAExpEvalList b nil nil
    | ABELCons : forall b e el n nl,
                 absAExpEval b e n ->
                 absAExpEvalList b el nl ->
                 absAExpEvalList b (e::el) (n::nl).
 
(* Predicate to combine a list of concrete heaps into a single heap *)  
                         
Inductive mergeHeapList : list heap -> heap -> Prop :=
    | MHNil : mergeHeapList nil empty_heap
    | MHList : forall f r hr h,
               mergeHeapList r hr ->
               disjoint f hr ->
               h = mergeHeaps f hr ->
               mergeHeapList (f::r) h.

(* Make a pointwise map between a list of ids and a list of nats *)

Fixpoint build_map (a : list id) (l : list nat) { struct l } : (id -> option nat) :=
    match l with
        | (f::r) => match a with
                        | (fa::ra) => fun x => (if beq_id x fa then Some f else (build_map ra r) x)
                        | nil => fun x => None
                    end
        | nil    => fun x => None
    end.


(* rMap is approximately the relation s, h |- \cal R_F(x,\overline X) in the paper.  
   The arguments line up as "rMap(h,F,x,\overline X,b)".  x is a nat (a variable value) 
   here unlike the paper where its a var; this gap is bridged below when s,h |- \Sigma defined. *)

Inductive rMap : heap -> (list id) -> nat -> list aexp -> (id -> (option nat)) -> Prop :=
    | RMStop : forall r e el b n,
               mem_exp e el=true ->
               absAExpEval b e n ->
               rMap empty_heap r n el b
    | RMNull : forall r vl b,
               rMap empty_heap r 0 vl b
    | RMStep : forall r n hl hr h vl b nl,
               sameLength r hl ->
               sameLength hl nl ->
               (forall s hh nn fmap, elementTriple s hh nn r hl nl ->
                                          Some fmap=h n -> (* note h here is h1 union .. union hn in paper *)
                                          Some nn = fmap s ->
                                          rMap hh r nn vl b) ->
               mergeHeapList hl h ->
               h n = None ->
               hr = (fun x => (if beq_nat x n then Some (build_map r nl) else h x)) ->
               rMap hr r n vl b.

Fixpoint fieldPresent (f : id) (map : list (id * aexp)) { struct map } : bool :=
    match map with
        | ((x,e)::b) => if beq_id x f then true else fieldPresent f b
        | nil => false
    end.

Fixpoint fieldValue (f : id) (map : list (id * aexp)) { struct map } : option aexp :=
    match map with
        | ((x,e)::b) => if beq_id x f then Some e else fieldValue f b
        | nil => None
    end.

Fixpoint replaceValue (map : list (id * aexp)) (f : id) (val : option aexp) : list (id * aexp) :=
   match val with
       | Some v =>
             match map with
                 | nil => ((f,v)::nil)
                 | ((f1,v1)::r) => if beq_id f f1 then (f,v)::r else (f1,v1)::(replaceValue r f (Some v))
             end
       | None =>
             match map with
                 | nil => nil
                 | ((f1,v1)::r) => if beq_id f f1 then r else (f1,v1)::(replaceValue r f None)
             end
    end.

Inductive selectCellField : (list (id * aexp)) -> (id * aexp) -> (list (id * aexp)) -> Prop :=
    | FselectCellField: forall f v r,
                        selectCellField ((f,v)::r) (f,v) r
    | RselectCellField: forall f v r e rr,
                        selectCellField r (f,v) rr ->
                        selectCellField (e::r) (f,v) (e::rr).

Inductive sameMap: (list (id * aexp)) -> (list (id * aexp)) -> Prop :=
    | sameMapLeft: forall f v l r rr,
                   sameMap l r ->
                   selectCellField rr (f,v) r ->
                   sameMap ((f,v)::l) rr
    | sameMapRight: forall f v l r rr,
                    sameMap r l ->
                    selectCellField rr (f,v) r ->
                    sameMap rr ((f,v)::l)
    | sameMapNil: sameMap nil nil.

(********************************************************************************
 *
 * The following predicate is used to map absHeaps from the abstract state
 * to concrete heaps.  The predicate takes three arguments, an absHeap, a
 * concrete heap and a set of variable bindings.  It is true when the concrete
 * heap is an instance of the abstract heap with the specified variable
 * bindings.  It corresponds to the predicate s, h |- \Sigma in the paper.
 *
 *******************************************************************************)

Inductive heapMap : absHeap -> heap -> binding -> Prop :=
    | CEEmpty : forall b, heapMap Empty empty_heap b
    | CECompose : forall h1 s1 h2 s2 b1 b2,
                  heapMap s1 h1 b1 ->
                  heapMap s2 h2 b2 ->
                  disjoint h1 h2 ->
                  compatible b1 b2 ->
                  heapMap (Compose s1 s2) (mergeHeaps h1 h2) (mergeBindings b1 b2)
    | CEImply : forall h1 s1 h2 h3 s3 b1 b2 b3,
                heapMap s1 h1 b1 ->
                heapMap s3 h3 b3 ->
                h3=mergeHeaps h1 h2 ->
                b3=mergeBindings b1 b2 ->
                disjoint h1 h2 ->
                compatible b1 b2 ->
                heapMap (HeapImply s1 s3) h2 b2
    | CESpatialCell : forall (h : heap) n map vmap b loc,
                      (forall x, not(x=n) -> (h x)=None) ->
                      absAExpEval b loc n ->
                      (h n)=Some map -> (* map is c in paper *)
                      (forall id v vv e, (map id = b v /\ not(b v=None)) -> 
                      (absAExpEval b e vv) -> b v=Some vv -> fieldValue id vmap = Some e) ->
                      heapMap (SpatialCell loc vmap) h b
    | CR : forall reg v n expList h b,
             rMap h reg n expList b ->
             absAExpEval b v n ->
             heapMap (R reg v expList) h b
    | CEJunk : forall h b,
               heapMap Junk h b.


(* This auxiliary definition indicates "s, h |- x \in \Sigma"; such a relation
   is not strictly defined in the paper. Arguments align with the paper as:
    memberOfSection(h,s,x,\Sigma) *)

Inductive memberOfSection : heap -> binding -> nat -> absHeap -> Prop :=
    | MSbind : forall h1 h2 b1 b2 n s fmap,
               heapMap s h1 b1 ->
               disjoint h1 h2 ->
               compatible b1 b2 ->
               h1 n = Some fmap ->
               memberOfSection (mergeHeaps h1 h2) (mergeBindings b1 b2) n s.
               
               
(*
 * heapCondPi below verifies the \Pi portion of an abstract state against the
 * heap component of a concrete state.  It is s, h |- \Pi in the paper.
 *)

(* First, we define heapCond, for the Body case of \Pi.  
   Different cases of this definition are given individualy in Section 3;
   here we define them in one definition. 
   The s, h |- Body of the paper aligns with heapCond(Body,true,h,s);
   s, h |/- Body then aligns with heapCond(Body,false,h,s)
*)


Inductive heapCond : formula -> bool -> heap -> binding -> Prop :=
    | FETrue : forall h b, heapCond FTrue true h b
    | FEFalse : forall h b, heapCond FFalse false h b
    | FEEqual: forall h b e1 e2 n,
               absAExpEval b e1 n ->
               absAExpEval b e2 n ->
               heapCond (FEqual e1 e2) true h b
    | FENEqual: forall h b e1 e2 n1 n2,
                absAExpEval b e1 n1 ->
                absAExpEval b e2 n2 ->
                beq_nat n1 n2=false ->
                heapCond (FEqual e1 e2) false h b
    | FELess : forall h b e1 e2 n1 n2,
               absAExpEval b e1 n1 ->
               absAExpEval b e2 n2 ->
	       n1 <= n2 ->
	       heapCond (FLessEqual e1 e2) true h b
    | FENLess1 : forall h b e1 e2 n1 n2,
                 absAExpEval b e1 n1 ->
                 absAExpEval b e2 n2 ->
                 n2 < n1 ->
                 heapCond (FLessEqual e1 e2) false h b
    | FEAnd : forall h b f1 f2,
              heapCond f1 true h b ->
              heapCond f2 true h b ->
              heapCond (FAnd f1 f2) true h b
    | FENAnd1 : forall h b f1 f2,
                heapCond f1 false h b ->
                heapCond (FAnd f1 f2) false h b
    | FENAnd2 : forall h b f1 f2,
                heapCond f2 true h b ->
                heapCond (FAnd f1 f2) false h b
    | FEOr1 : forall h b f1 f2,
              heapCond f1 true h b ->
              heapCond (FOr f1 f2) true h b
    | FEOr2 : forall h b f1 f2,
              heapCond f2 true h b ->
              heapCond (FOr f1 f2) true h b
    | FENOr1 : forall h b f1 f2,
               heapCond f1 false h b ->
               heapCond f2 false h b ->
               heapCond (FOr f1 f2) false h b
    | FENot : forall h b f,
              heapCond f false h b ->
              heapCond (FNot f) true h b
    | FENNot : forall h b f,
               heapCond f true h b ->
               heapCond (FNot f) false h b
    | FEAll  : forall v f b h set,
               (forall n,
                       memberOfSection h b n set ->
                       heapCond f true h (addBinding b v n)) ->
               heapCond (FAll v set f) true h b
    | FENAll : forall v f n b h set,
               memberOfSection h b n set ->
               heapCond f false h (addBinding b v n) ->
               heapCond (FAll v set f) false h b
    | FEExists : forall v f b h n set,
                 memberOfSection h b n set ->
                 heapCond f true h (addBinding b v n) ->
                 heapCond (FExists v set f) true h b
    | FENExists : forall v f b h set,
                  (forall n, memberOfSection h b n set ->
                             heapCond f false h (addBinding b v n)) ->
                  heapCond (FExists v set f) false h b
    (*| FLeadsTo : forall h b v1 v2 r n1 n2 el nl,
                 absAExpEval b v1 n1 ->
                 absAExpEval b v2 n2 ->
                 absAExpEvalList b el nl ->
                 pointsTo h r n1 n2 nl ->
                 heapCond (LeadsTo v1 r el v2) true h b*)
    | FERef : forall h b v1 f n1 v2 n2 fmap,
              absAExpEval b v1 n1 ->
              absAExpEval b v2 n2 ->
              (h n1) = Some fmap ->
              fmap f = Some n2 ->
              heapCond (LeadsTo v1 f v2) true h b.

Inductive heapCondPi : pi -> bool -> heap -> binding -> Prop :=
    | FMember1: forall v set h b f n,
                (b v)=None ->
                memberOfSection h b n set ->
                heapCondPi f true h (addBinding b v n) ->
                heapCondPi (PMember v set f) true h b
    | FMember2: forall v set h b f n,
                (b v)=None ->
                not(memberOfSection h b n set) ->
                heapCondPi (PMember v set f) false h b
    | FMember3: forall v set h b f n,
                (b v)=None ->
                memberOfSection h b n set ->
                heapCondPi f false h (addBinding b v n) ->
                heapCondPi (PMember v set f) false h b
    | FMember4: forall v set h b f,
                (b v)=None ->
                heapCondPi f false h (addBinding b v 0) ->
                heapCondPi (PMember v set f) false h b
    | FBody: forall b h f c,
             heapCond f c h b ->
             heapCondPi (PBody f) c h b.

(* ***************************************************************************
 *
 * stateMap is the top level predicate mapping abstract states to concrete
 * states.  It corresponds to h, s |- \Pi | \Sigma of the paper.  Note that the
 * "exists p ..." clause is there to give the primed variables their proper
 * existential semantics.
 *
 *****************************************************************************)

Inductive stateMap : absState -> state -> Prop :=
    | FEMapping : forall h e m f b,
                  (exists p,(forall n, p (Id n)=None) ->
                            heapMap m h (fun x => match p x with None => b x | x => x end) ->
                            heapCondPi f true h (fun x => match b x with None => b x | x => x end)) ->
                  (forall n x, b (Id n)=Some x -> x=(e (Id n))) ->
                  (forall n, b (PrimedId n)=None) ->
                  stateMap (m,f) (e,h).

(* ***************************************************************************
 *
 * Here are definitions of equivalent formulas and spatial maps
 *
 * *************************************************************************)

Definition equivalent_formulas (f1 : formula) (f2 : formula) : Prop :=
    forall b t h,
        heapCond f1 t h b -> heapCond f2 t h b /\
        heapCond f2 t h b -> heapCond f1 t h b.

Inductive equivalent_spatial: absHeap -> absHeap -> Prop :=
    | SBasic : forall b h s1 s2,
       heapMap s1 h b -> heapMap s2 h b ->
       heapMap s2 h b -> heapMap s1 h b ->
       equivalent_spatial s1 s2.

Definition equivalent_pi (p1 : pi) (p2 : pi) : Prop :=
    forall b t h,
       heapCondPi p1 t h b -> heapCondPi p2 t h b /\
       heapCondPi p2 t h b -> heapCondPi p1 t h b.

(* ***************************************************************************
 *
 * The var_free predicates below check whether or not a variable is free in a 
 * expression/abstract heap/formula/Pi
 *
 *****************************************************************************)

Inductive var_free_exp : aexp -> bool -> id -> Prop :=
    | VFANum : forall x id, var_free_exp (ANum x) false id
    | VFAVar1 : forall x id,
                (beq_id id x)=true ->
                var_free_exp (AVar x) true id
    | VFAVar2 : forall x id,
                (beq_id id x)=false ->
                var_free_exp (AVar x) false id
    | APlus1 : forall e1 e2 id,
               var_free_exp e1 true id -> var_free_exp (APlus e1 e2) true id
    | APlus2 : forall e1 e2 id,
               var_free_exp e2 true id -> var_free_exp (APlus e1 e2) true id
    | APlus3 : forall e1 e2 id,
               var_free_exp e2 false id ->
               var_free_exp e1 false id ->
               var_free_exp (APlus e1 e2) false id
    | AMinus1 : forall e1 e2 id,
               var_free_exp e1 true id -> var_free_exp (AMinus e1 e2) true id
    | AMinus2 : forall e1 e2 id,
               var_free_exp e2 true id -> var_free_exp (AMinus e1 e2) true id
    | AMinus3 : forall e1 e2 id,
               var_free_exp e2 false id ->
               var_free_exp e1 false id ->
               var_free_exp (AMinus e1 e2) false id
    | AMult1 : forall e1 e2 id,
               var_free_exp e1 true id -> var_free_exp (AMult e1 e2) true id
    | AMult2 : forall e1 e2 id,
               var_free_exp e2 true id -> var_free_exp (AMult e1 e2) true id
    | AMult3 : forall e1 e2 id,
               var_free_exp e2 false id ->
               var_free_exp e1 false id ->
               var_free_exp (AMult e1 e2) false id.

Inductive var_free_exp_list : (list aexp) -> bool -> id -> Prop :=
    | VFL1 : forall e r id, var_free_exp e true id -> var_free_exp_list (e::r) true id
    | VFL2 : forall e r id, var_free_exp_list r true id -> var_free_exp_list (e::r) true id
    | VFL3 : forall e r id, var_free_exp e false id ->
                            var_free_exp_list r false id ->
                            var_free_exp_list (e::r) false id
    | VFL4 : forall false id, var_free_exp_list nil false id.


Inductive var_free_spatial : absHeap -> bool -> id -> Prop :=
    | VFSEmpty : forall n, var_free_spatial Empty true n
    | VFSCompose1 : forall h1 h2 n,
                    var_free_spatial h1 true n -> var_free_spatial (Compose h1 h2) true n
    | VFSCompose2 : forall h1 h2 n,
                    var_free_spatial h2 true n -> var_free_spatial (Compose h1 h2) true n
    | VFSCompose3 : forall h1 h2 n,
                    var_free_spatial h1 false n ->
                    var_free_spatial h2 false n ->
                    var_free_spatial (Compose h1 h2) false n
    | VFSHeapImply1 : forall h1 h2 n,
                      var_free_spatial h1 true n -> var_free_spatial (HeapImply h1 h2) true n
    | VFSHeapImply2 : forall h1 h2 n,
                      var_free_spatial h2 true n -> var_free_spatial (HeapImply h1 h2) true n
    | VFSHeapImply3 : forall h1 h2 n,
                      var_free_spatial h1 false n ->
                      var_free_spatial h2 false n ->
                      var_free_spatial (HeapImply h1 h2) false n
    | VFSSpatialCell1: forall fmap v l,
                       var_free_exp l true v -> var_free_spatial (SpatialCell l fmap) true v
    | VFSSpatialCell2: forall fmap v l e f,
                       fieldValue f fmap = Some e ->
                       var_free_exp e true v ->
                       var_free_spatial (SpatialCell l fmap) true v
    | VFSSpatialCell3: forall fmap v l,
                       var_free_exp l false v ->
                       (forall e f, (fieldValue f fmap=None \/
                                     ((fieldValue f fmap=Some e ->
                                       var_free_exp e false v)))) ->
                       var_free_spatial (SpatialCell l fmap) false v
    | VFSR1 : forall rr vv lvv v,
              var_free_exp vv true v ->
              var_free_spatial (R rr vv lvv) true v
    | VFSR2 : forall rr vv lvv v,
              var_free_exp_list lvv true v ->
              var_free_spatial (R rr vv lvv) true v
    | VFSR3 : forall rr vv lvv v,
              var_free_exp_list lvv false v ->
              var_free_exp vv false v ->
              var_free_spatial (R rr vv lvv) false v
    | VFSJunk : forall v, var_free_spatial Junk false v.


Inductive var_free_formula : formula -> bool -> id -> Prop :=
    | VFFTrue : forall v, var_free_formula FTrue false v
    | VFFFalse : forall v, var_free_formula FFalse false v
    | VFFEqual1 : forall v e1 e2,
                  var_free_exp e1 true v ->
                  var_free_formula (FEqual e1 e2) true v
    | VFFEqual2 : forall v e1 e2,
                  var_free_exp e2 true v ->
                  var_free_formula (FEqual e1 e2) true v
    | VFFEqual3 : forall v e1 e2,
                  var_free_exp e1 false v ->
                  var_free_exp e2 false v ->
                  var_free_formula (FEqual e1 e2) false v
    | VFFLessEqual1 : forall v e1 e2,
                      var_free_exp e1 true v ->
                      var_free_formula (FLessEqual e1 e2) true v
    | VFFLessEqual2 : forall v e1 e2,
                      var_free_exp e2 true v ->
                      var_free_formula (FLessEqual e1 e2) true v
    | VFFLessEqual3 : forall v e1 e2,
                      var_free_exp e1 false v ->
                      var_free_exp e2 false v ->
                      var_free_formula (FLessEqual e1 e2) false v
    | VFFAnd1 : forall v f1 f2,
                var_free_formula f1 true v ->
                var_free_formula (FAnd f1 f2) true v
    | VFFAnd2 : forall v f1 f2,
                var_free_formula f2 true v ->
                var_free_formula (FAnd f1 f2) true v
    | VFFAnd3 : forall v f1 f2,
                var_free_formula f1 false v ->
                var_free_formula f2 false v ->
                var_free_formula (FAnd f1 f2) false v
    | VFFOr1 : forall v f1 f2,
               var_free_formula f1 true v ->
               var_free_formula (FOr f1 f2) true v
    | VFFOr2 : forall v f1 f2,
               var_free_formula f2 true v ->
               var_free_formula (FOr f1 f2) true v
    | VFFOr3 : forall v f1 f2,
               var_free_formula f1 false v ->
               var_free_formula f2 false v ->
               var_free_formula (FOr f1 f2) false v
    | VFFNot1 : forall v f,
                var_free_formula f true v ->
                var_free_formula (FNot f) true v
    | VFFNot2 : forall v f,
                var_free_formula f false v ->
                var_free_formula (FNot f) false v
    | VFFAll1: forall v set f,
              var_free_formula (FAll v set f) true v
    | VFFAll2: forall v var set f,
               var_free_spatial set false v ->
               var_free_formula f false v ->
               var_free_formula (FAll var set f) false v
    | VFFAll3: forall v var set f,
               var_free_spatial set true v ->
               var_free_formula (FAll var set f) true v
    | VFFAll4: forall v var set f,
               (beq_id v var)=false ->
               var_free_formula f true v ->
               var_free_formula (FAll var set f) true v
    | VFFExists1: forall v set f,
                  var_free_formula (FExists v set f) true v
    | VFFExists2: forall v var set f,
                  var_free_spatial set false v ->
                  var_free_formula f false v ->
                  var_free_formula (FExists var set f) false v
    | VFFExists3: forall v var set f,
                  var_free_spatial set true v ->
                  var_free_formula (FExists var set f) true v
    | VFFExists4: forall v var set f,
                  (beq_id v var)=false ->
                  var_free_formula f true v ->
                  var_free_formula (FExists var set f) true v
    | VFFLeadsTo1: forall s f t v,
                   var_free_exp s true v ->
                   var_free_formula (LeadsTo s f t) true v
    | VFFLeadsTo2: forall s f t v,
                   var_free_exp t true v ->
                   var_free_formula (LeadsTo s f t) true v
    | VFFLeadsTo3: forall s f t v,
                   var_free_exp t false v ->
                   var_free_exp s false v ->
                   var_free_formula (LeadsTo s f t) false v.


Inductive var_free_pi : pi -> bool -> id -> Prop :=
    | VFPMember1: forall v set pi,
                  var_free_pi (PMember v set pi) true v
    | VFPMember2: forall v set pi var,
                  var_free_spatial set true var ->
                  var_free_pi (PMember v set pi) true var
    | VFPMember3: forall v set pi var,
                  var_free_pi pi true var ->
                  var_free_pi (PMember v set pi) true var
    | VFPMember4: forall v set pi var,
                  (beq_id v var)=false ->
                  var_free_spatial set false var ->
                  var_free_pi pi false var ->
                  var_free_pi (PMember v set pi) false var
    | VFPBody:    forall f c v,
                  var_free_formula f c v ->
                  var_free_pi (PBody f) c v.

Hint Resolve VFANum VFAVar1 VFAVar2 APlus1 APlus2 APlus3 AMinus1 AMinus2 AMinus3
             AMult1 AMult2 AMult3
             VFL1 VFL2 VFL3 VFL4
             VFSEmpty VFSCompose1 VFSCompose2 VFSCompose3 VFSHeapImply1 VFSHeapImply2
             VFSHeapImply3 VFSSpatialCell1 VFSSpatialCell2 VFSSpatialCell3 VFSR1
             VFSR2 VFSR3 VFSJunk
             VFFTrue VFFFalse VFFEqual1 VFFEqual2 VFFEqual3 VFFLessEqual1 VFFLessEqual2
             VFFLessEqual3 VFFAnd1 VFFAnd2 VFFAnd3 VFFOr1 VFFOr2 VFFOr3 VFFNot1 VFFNot2
             VFFAll1 VFFAll2 VFFAll3 VFFAll4 VFFExists1 VFFExists2 VFFExists3 VFFExists4
             VFFLeadsTo1 VFFLeadsTo2 VFFLeadsTo3
             VFPMember1 VFPMember2 VFPMember3 VFPMember4 VFPBody : varFree.

Ltac reduceVarFree := (apply VFSCompose3; try reduceVarFree) ||
                      (apply VFPMember4; try reduceVarFree) ||
                      (apply VFPBody; try reduceVarFree) ||
                      (apply VFFAll2; try reduceVarFree) ||
                      (apply VFFExists2; try reduceVarFree) ||
                      (apply VFFOr1; try reduceVarFree) ||
                      (apply VFFOr2; try reduceVarFree) ||
                      (apply VFFOr3; try reduceVarFree) ||
                      (apply VFFAnd1; try reduceVarFree) ||
                      (apply VFFAnd2; try reduceVarFree) ||
                      (apply VFFAnd3; try reduceVarFree) ||
                      (apply VFFNot1; try reduceVarFree) ||
                      (apply VFFNot2; try reduceVarFree) ||
                      (apply VFSSpatialCell3; try reduceVarFree) ||
                      simpl ||
                      (repeat auto with varFree).

(* *************************************************************************
 *
 * Simplification: basic logical simplification
 * These principles should not be hard to prove sound
 *
 * *************************************************************************)


(* *************************************************************************
 * Here are some logical axioms for equivalent pi's and formulas.  They have been
 * carefully structured so as to be convienent to chain together in a
 * complex equivalency proof.
 *
 * *************************************************************************)

Theorem equivalent_pi_reflexive :
    forall f,
        equivalent_pi f f.
Admitted.
Hint Resolve equivalent_pi_reflexive.

Theorem equivalent_pi_pmember1:
    forall p1 p2 start v set,
        equivalent_pi p1 p2 ->
        equivalent_pi start (PMember v set p1) ->
        equivalent_pi start (PMember v set p2).
Admitted.
Hint Resolve equivalent_pi_pmember1.

Theorem equivalent_pi_pmember_elim:
    forall start v set p,
        var_free_pi p false v ->
        equivalent_pi start p ->
        equivalent_pi start (PMember v set p).
Admitted.

Theorem equivalent_pi_pbody:
    forall f1 f2 start,
        equivalent_formulas f1 f2 ->
        equivalent_pi start (PBody f1) ->
        equivalent_pi start (PBody f2).
Admitted.


Theorem equivalent_formula_reflexive :
    forall f,
        equivalent_formulas f f.

Admitted.
Hint Resolve equivalent_formula_reflexive.

Theorem equivalent_formula_and1:
    forall r f1 f2 start,
        equivalent_formulas f1 f2 ->
        equivalent_formulas start (FAnd f1 r) ->
        equivalent_formulas start (FAnd f2 r).

Admitted.
Hint Resolve equivalent_formula_and1.

Theorem equivalent_formula_and2:
    forall r f1 f2 start,
        equivalent_formulas f1 f2 ->
        equivalent_formulas start (FAnd r f1) ->
        equivalent_formulas start (FAnd r f2).

Admitted.
Hint Resolve equivalent_formula_and2.

Theorem equivalent_formula_or1:
    forall r f1 f2 start,
        equivalent_formulas f1 f2 ->
        equivalent_formulas start (FOr f1 r) ->
        equivalent_formulas start (FOr f2 r).

Admitted.
Hint Resolve equivalent_formula_or1.

Theorem equivalent_formula_or2:
    forall r f1 f2 start,
        equivalent_formulas f1 f2 ->
        equivalent_formulas start (FOr r f1) ->
        equivalent_formulas start (FOr r f2).

Admitted.
Hint Resolve equivalent_formula_or2.

Theorem equivalent_formula_not:
    forall f1 f2 start,
        equivalent_formulas f1 f2 ->
        equivalent_formulas start (FNot f1) ->
        equivalent_formulas start (FNot f2).

Admitted.
Hint Resolve equivalent_formula_not.

Theorem equivalent_formula_exists:
    forall f1 f2 start v set,
        equivalent_formulas f1 f2 ->
        equivalent_formulas start (FExists v set f1) ->
        equivalent_formulas start (FExists v set f2).

Admitted.
Hint Resolve equivalent_formula_exists.

Theorem equivalent_formula_exists_set:
    forall set1 set2 f start v,
        equivalent_spatial set1 set2 ->
        equivalent_formulas start (FExists v set1 f) ->
        equivalent_formulas start (FExists v set2 f).
Admitted.
Hint Resolve equivalent_formula_exists_set.

Theorem equivalent_formula_all:
    forall f1 f2 start v set,
        equivalent_formulas f1 f2 ->
        equivalent_formulas start (FAll v set f1) ->
        equivalent_formulas start (FAll v set f2).

Admitted.
Hint Resolve equivalent_formula_all.

Theorem equivalent_formula_all_set:
    forall set1 set2 f start v,
        equivalent_spatial set1 set2 ->
        equivalent_formulas start (FAll v set1 f) ->
        equivalent_formulas start (FAll v set2 f).
Admitted.
Hint Resolve equivalent_formula_all_set.

(*Theorem equivalent_formula_self_pointer:
    forall e r start,
        equivalent_formulas start FTrue ->
        equivalent_formulas start (r ~ (Star e) ~> r).

Admitted.
Hint Resolve equivalent_formula_self_pointer.*)

(*Theorem equivalent_formula_pointer_null_source:
    forall e r start,
        equivalent_formulas start FFalse ->
        equivalent_formulas start (A0 ~ e ~> r).

Admitted.
Hint Resolve equivalent_formula_pointer_null_source.*)

Theorem equivalent_formula_and_true1:
    forall e start,
        equivalent_formulas start e ->
        equivalent_formulas start (FAnd FTrue e).

Admitted.
Hint Resolve equivalent_formula_and_true1.

Theorem equivalent_formula_and_true2:
    forall e start,
        equivalent_formulas start e ->
        equivalent_formulas start (FAnd e FTrue).

Admitted.
Hint Resolve equivalent_formula_and_true2.

Theorem equivalent_formula_and_false1:
    forall e start,
        equivalent_formulas start FFalse ->
        equivalent_formulas start (FAnd FFalse e).

Admitted.
Hint Resolve equivalent_formula_and_false1.

Theorem equivalent_formula_and_false2:
    forall e start,
        equivalent_formulas start FFalse ->
        equivalent_formulas start (FAnd e FFalse).

Admitted.
Hint Resolve equivalent_formula_and_false2.

Theorem equivalent_formula_or_true1:
    forall e start,
        equivalent_formulas start FTrue ->
        equivalent_formulas start (FOr FTrue e).

Admitted.
Hint Resolve equivalent_formula_and_true1.

Theorem equivalent_formula_or_true2:
    forall e start,
        equivalent_formulas start FTrue ->
        equivalent_formulas start (FOr e FTrue).

Admitted.
Hint Resolve equivalent_formula_and_true2.

Theorem equivalent_formula_or_false1:
    forall e start,
        equivalent_formulas start e ->
        equivalent_formulas start (FOr FFalse e).

Admitted.
Hint Resolve equivalent_formula_or_false1.

Theorem equivalent_formula_or_false2:
    forall e start,
        equivalent_formulas start e ->
        equivalent_formulas start (FOr e FFalse).

Admitted.
Hint Resolve equivalent_formula_or_false2.

Theorem equivalent_formula_not_false:
    forall start,
        equivalent_formulas start FTrue ->
        equivalent_formulas start (FNot FFalse).

Admitted.
Hint Resolve equivalent_formula_not_false.

Theorem equivalent_formula_not_true:
    forall start,
        equivalent_formulas start FFalse ->
        equivalent_formulas start (FNot FTrue).

Admitted.
Hint Resolve equivalent_formula_not_true.

Theorem equivalent_formula_all_true:
    forall start v set,
        equivalent_formulas start FTrue ->
        equivalent_formulas start (FAll v set FTrue).

Admitted.
Hint Resolve equivalent_formula_all_true.

Theorem equivalent_formula_all_true_empty:
    forall start v f,
        equivalent_formulas start FTrue ->
        equivalent_formulas start (FAll v Empty f).
Admitted.
Hint Resolve equivalent_formula_all_true_empty.

Theorem equivalent_formula_all_false:
    forall start v set,
        equivalent_formulas start FFalse ->
        equivalent_formulas start (FAll v set FFalse).

Admitted.
Hint Resolve equivalent_formula_all_false.

Theorem equivalent_formula_exists_true:
    forall start v set,
        equivalent_formulas start FTrue ->
        equivalent_formulas start (FExists v set FTrue).

Admitted.
Hint Resolve equivalent_formula_exists_true.

Theorem equivalent_formula_exists_false:
    forall start v set,
        equivalent_formulas start FFalse ->
        equivalent_formulas start (FExists v set FFalse).

Admitted.
Hint Resolve equivalent_formula_exists_false.

(* *************************************************************************
 *
 * Here are some axioms about equivalent spatial expressions.  They have
 * been carefully structured so as to be convienent to chain together in a
 * complex equivalency proof.
 *
 * *************************************************************************)

Theorem equivalent_spatial_reflexive :
    forall l,
		equivalent_spatial l l.

Admitted.
Hint Resolve equivalent_spatial_reflexive.

Theorem equivalent_spatial_compose1 :
    forall r l1 l2 start,
        equivalent_spatial l1 l2 ->
        equivalent_spatial start (Compose l1 r) ->
        equivalent_spatial start (Compose l2 r).

Admitted.
Hint Resolve equivalent_spatial_compose1.

Theorem equivalent_spatial_compose2 :
    forall r1 r2 l start,
        equivalent_spatial r1 r2 ->
        equivalent_spatial start (Compose l r1) ->
        equivalent_spatial start (Compose l r2).

Admitted.
Hint Resolve equivalent_spatial_compose2.

Theorem equivalent_empty_R :
    forall el r start,
        equivalent_spatial start Empty ->
        equivalent_spatial start (R r (ANum 0) el).

Admitted.
Hint Resolve equivalent_empty_R.

Theorem equivalent_empty_compose1 :
    forall s start,
        equivalent_spatial start s ->
        equivalent_spatial start (Compose Empty s).

Admitted.
Hint Resolve equivalent_empty_compose1.

Theorem equivalent_empty_compose2 :
    forall s start,
        equivalent_spatial start s ->
        equivalent_spatial start (Compose s Empty).

Admitted.
Hint Resolve equivalent_empty_compose2.

(* basic conjunction of logical constante simplification, e.g.  
   simplifies logical formula true and false to just false *)

Fixpoint andSimp (f : formula) : formula :=
    match f with
        | (FAnd f1 f2) => match (andSimp f1) with
                              | FFalse => FFalse
                              | FTrue => (andSimp f2)
                              | x => match (andSimp f2) with
                                        | FFalse => FFalse
                                        | FTrue => x
                                        | y => (FAnd x y)
                                    end
                         end
        | x => x
    end.

(* axiom to apply the above simpification *)

Theorem equivalent_formula_and_simp :
    forall s start,
        equivalent_formulas start (andSimp s) ->
        equivalent_formulas start s.

Admitted.
Hint Resolve equivalent_formula_and_simp.

(* *************************************************************************
 * This completes the axioms for equivalent formulae.  Now we give
 * analogous logical axioms for strengthening formulae and abstract heaps. 
 * consequent_formula(f1,f2) means f2 implies f1 in all contexts.
 *
 * *************************************************************************)

Inductive consequent_spatial: absHeap -> absHeap -> Prop :=
    | CSBasic : forall s1 s2 b h,
      heapMap s2 h b -> heapMap s1 h b ->
      consequent_spatial s1 s2.

Inductive consequent_formula : formula -> formula -> Prop :=
    | CFBasic : forall f1 f2 b t h,
        heapCond f2 t h b -> heapCond f1 t h b ->
        consequent_formula f1 f2.

Inductive consequent_pi : pi -> pi -> Prop :=
    | CFBasicPi : forall f1 f2 b t h,
        heapCondPi f2 t h b -> heapCondPi f1 t h b ->
        consequent_pi f1 f2.

Theorem consequent_formula_reflexive :
    forall f,
        consequent_formula f f.

Admitted.
Hint Resolve consequent_formula_reflexive.

Theorem consequent_formula_cancel_and1 :
    forall s start f,
        consequent_formula start s ->
        consequent_formula start (FAnd s f).

Admitted.
Hint Resolve consequent_formula_cancel_and1.

Theorem consequent_formula_cancel_and2 :
    forall s start f,
        consequent_formula start s ->
        consequent_formula start (FAnd f s).

Admitted.
Hint Resolve consequent_formula_cancel_and2.

Fixpoint addFormula (p : pi) (f : formula) { struct p } : pi :=
    match p with
        | PMember v set x => PMember v set (addFormula x f)
        | PBody x => PBody (FAnd x f)
    end.

(* This is the formula strengthening axiom: *)

Theorem consequent_pi_add_formula:
    forall start f pi,
        consequent_pi start pi ->
        consequent_pi start (addFormula pi f).

Admitted.
Hint Resolve consequent_pi_add_formula.

Theorem consequent_pi_reflexive:
    forall pi,
        consequent_pi pi pi.
Admitted.
Hint Resolve consequent_pi_reflexive.

Theorem consequent_pi_pmember:
    forall pi v s start,
        consequent_pi start pi ->
        consequent_pi start (PMember v s pi).
Admitted.
Hint Resolve consequent_pi_pmember.

Theorem consequent_pi_left_pmember:
    forall pi v s finish,
        consequent_pi (PMember v s pi) finish ->
        consequent_pi pi finish.
Admitted.
Hint Resolve consequent_pi_left_pmember.

Theorem consequent_pi_both_pmember:
    forall pi v s finish,
        consequent_pi pi finish ->
        consequent_pi (PMember v s pi) (PMember v s finish).
Admitted.
Hint Resolve consequent_pi_both_pmember.

Theorem consequent_pi_pbody:
    forall l r,
        consequent_formula l r ->
        consequent_pi (PBody l) (PBody r).
Admitted.
Hint Resolve consequent_pi_pbody.

Theorem consequent_pi_pbody_true:
    forall x,
        consequent_pi (PBody FTrue) x.
Admitted.
Hint Resolve consequent_pi_pbody_true.

(* Two auxiliary functions to extract and replace the Body in a \Pi *)

Fixpoint extractBody (p : pi) { struct p } : formula :=
    match p with
        | PMember v set p => extractBody p
        | PBody f => f
    end.

Fixpoint replaceBody (p : pi) (b : formula) {struct p } : pi :=
    match p with
        | PMember v set p => (PMember v set (replaceBody p b))
        | PBody _ => (PBody b)
    end.

(* Auxiliary definition to weaken a formula by pulling out a conjunction
   For andComponent(a,b,c), logically a and c iff b *)

Inductive andComponent: formula -> formula -> formula -> Prop :=
    | ACTrue : andComponent FTrue FTrue FTrue
    | ACFalse : andComponent FFalse FFalse FTrue
    | ACEqual: forall e1 e2, andComponent (FEqual e1 e2) (FEqual e1 e2) FTrue
    | ACLessEqual : forall e1 e2, andComponent (FLessEqual e1 e2) (FLessEqual e1 e2) FTrue
    | ACAnd1 : forall  e f1 f2 x,
              andComponent e f1 x ->
              andComponent e (FAnd f1 f2) (FAnd f2 x)
    | ACAnd2 : forall  e f1 f2 x,
              andComponent e f2 x ->
              andComponent e (FAnd f1 f2) (FAnd x f1)
    | ACOr : forall f1 f2, andComponent (FOr f1 f2) (FOr f1 f2) FTrue
    | ACNot : forall f, andComponent (FNot f) (FNot f) FTrue
    | ACAll  : forall v f set, andComponent (FAll v set f) (FAll v set f) FTrue
    | ACExists : forall v f set, andComponent (FExists v set f) (FExists v set f) FTrue
    | ACRef : forall v1 f v2, andComponent (LeadsTo v1 f v2) (LeadsTo v1 f v2) FTrue.

(* weaken a pi by removing arbitrary conjunctions at the outside of the formula inside *)

Theorem consequent_pi_and:
    forall pi finish b,
        b = extractBody pi ->
        (forall c x, andComponent c b x -> consequent_pi (replaceBody  pi c) finish) ->
        consequent_pi pi finish.
Admitted.
Hint Resolve consequent_pi_and.

(* as above but on the formula itself *)

Theorem consequent_formula_and:
    forall b finish,
        (forall c x, andComponent c b x -> consequent_formula c finish) ->
        consequent_formula b finish.
Admitted.
Hint Resolve consequent_formula_and.

(* weaken a formula or a pi by removing a selected conjunction x *)

Theorem consequent_formula_select:
    forall b c finish x,
       consequent_formula b c ->
       andComponent c finish x ->
       consequent_formula b finish.
Admitted.
Hint Resolve consequent_formula_select.

Theorem consequent_pi_select:
    forall b c finish pi x,
       b = extractBody finish ->
       consequent_pi pi (replaceBody finish c) ->
       andComponent c b x ->
       consequent_pi pi finish.
Admitted.
Hint Resolve consequent_pi_select.

(* Specialized tactics for formula consequent simplification *)

Ltac consequent_select_recurse :=
     solve [ (apply ACAnd1; consequent_select_recurse) |
             (apply ACAnd2; consequent_select_recurse) |
             apply ACTrue |
             apply ACFalse |
             apply ACEqual |
             apply ACLessEqual |
             apply ACOr |
             apply ACNot |
             apply ACAll |
             apply ACExists |
             apply ACRef].

Ltac consequent_pi_pmember_elim := (apply consequent_pi_pmember; consequent_pi_pmember_elim) || simpl.

Ltac consequent_pi_select_elim :=
     simpl; eapply consequent_pi_select; [simpl; reflexivity | simpl; consequent_pi_pmember_elim; apply consequent_pi_reflexive | consequent_select_recurse].

Ltac consequent_formula_select_elim :=
     simpl; eapply consequent_formula_select; [simpl; apply consequent_formula_reflexive | consequent_select_recurse].
     
Ltac consequent_formula_and_inversion:=
     match goal with
     | id: andComponent ?V ?F ?R |- _ => inversion id; subst; clear id
     end; (try consequent_formula_and_inversion || simpl).
(*Ltac consequent_formula_and :=
     eapply consequent_formula_and; simpl; reflexivity*)

Theorem consequent_spatial_reflexive:
    forall h, consequent_spatial h h.
Admitted.
Hint Resolve consequent_spatial_reflexive.

(* Simplification for logical negations.
   Push a negation into formula f via DeMorgans laws; 
   neg is the co/contravariance flag *)

Fixpoint reduceNot (f : formula) (neg : bool) { struct f } : formula :=
    if neg then
        match f with
        | FTrue => FFalse
        | FFalse => FTrue
        | (FEqual e1 e2) => (FNot (FEqual e1 e2))
        | (FLessEqual e1 e2) => (FNot (FLessEqual e1 e2))
        | (FAnd f1 f2) => (FOr (reduceNot f1 true) (reduceNot f2 true))
        | (FOr f1 f2) => (FOr (reduceNot f1 true) (reduceNot f2 true))
        | (FNot f) => reduceNot f false
        | (FAll v set f) => (FExists v set (reduceNot f true))
        | (FExists v set f) => (FAll v set (reduceNot f true))
        | (LeadsTo v1 f v2) => (FNot (LeadsTo v1 f v2))
        end
    else
        match f with
        | FTrue => FTrue
        | FFalse => FFalse
        | (FEqual e1 e2) => (FEqual e1 e2)
        | (FLessEqual e1 e2) => (FLessEqual e1 e2)
        | (FAnd f1 f2) => (FAnd (reduceNot f1 false) (reduceNot f2 false))
        | (FOr f1 f2) => (FOr (reduceNot f1 false) (reduceNot f2 false))
        | (FNot f) => reduceNot f true
        | (FAll v set f) => (FAll v set (reduceNot f false))
        | (FExists v set f) => (FExists v set (reduceNot f false))
        | (LeadsTo v1 f v2) => (LeadsTo v1 f v2)
        end.

(* Axiom to apply the above not reduction *)

Theorem consequent_formula_reduce_not:
     forall l r,
        consequent_formula (reduceNot l false) (reduceNot r false) ->
        consequent_formula l r.
Admitted.
Hint Resolve consequent_formula_reduce_not.

(* Simplification principles for logical disjunctions *)

Inductive orComponent: formula -> formula -> formula -> Prop :=
    | OCTrue : orComponent FTrue FTrue FFalse
    | OCFalse : orComponent FFalse FFalse FFalse
    | OCEqual: forall e1 e2, orComponent (FEqual e1 e2) (FEqual e1 e2) FFalse
    | OCLessEqual : forall e1 e2, orComponent (FLessEqual e1 e2) (FLessEqual e1 e2) FFalse
    | OCOr1 : forall  e f1 f2 x,
              orComponent e f1 x ->
              orComponent e (FOr f1 f2) (FOr f2 x)
    | OCOr2 : forall  e f1 f2 x,
              orComponent e f2 x ->
              orComponent e (FOr f1 f2) (FOr x f1)
    | OCAnd : forall f1 f2, orComponent (FAnd f1 f2) (FAnd f1 f2) FFalse
    | OCNot : forall f, orComponent (FNot f) (FNot f) FFalse
    | OCAll  : forall v f set, orComponent (FAll v set f) (FAll v set f) FFalse
    | OCExists : forall v f set, orComponent (FExists v set f) (FExists v set f) FFalse
    | OCRef : forall v1 f v2, orComponent (LeadsTo v1 f v2) (LeadsTo v1 f v2) FFalse.

Fixpoint orSimp (f : formula) : formula :=
    match f with
        | (FOr f1 f2) => match (orSimp f1) with
                             | FTrue => FTrue
                             | FFalse => (orSimp f2)
                             | x => match (orSimp f2) with
                                        | FTrue => FTrue
                                        | FFalse => x
                                        | y => (FOr x y)
                                    end
                         end
        | x => x
    end.

Inductive resolve: formula -> formula -> Prop :=
     Resolve: forall f c1 f1 c2 f2 fn fr fsimp,
              andComponent f1 f c1 ->
              andComponent f2 c1 c2 ->
              fn = reduceNot f1 true ->
              orComponent fn f2 fr ->
              fsimp = orSimp fr ->
              resolve f (FAnd c2 fsimp).

Theorem consequent_formula_resolve2:
    forall l r1 r2,
       resolve r1 r2 ->
       consequent_formula l r2 ->
       consequent_formula l r1.
Admitted.


(* ******************************************************
 *
 * Now we present several misc. axioms which are needed in the proof.
 *
 ********************************************************)

(* Axiom for removing duplicate Pmember clauses at the head of a \Pi. *)

(* Auxiliary function removePmember(v,set,pi1,p2) iff pi2 is pi1 with a "v in set" clause removed *)

Inductive removePmember: id -> absHeap -> pi -> pi -> Prop :=
    | RemovePmember: forall v set b,
          removePmember v set (PMember v set b) b
    | RemoveNextPmember: forall v set v2 set2 b b2,
          removePmember v set b b2 ->
          removePmember v set (PMember v2 set2 b) (PMember v2 set2 b2).

(* remove_dup_pmember(pi1,pi2) iff pi2 is pi1 with a duplicate pmember clause removed *)

Inductive remove_dup_pmember: pi -> pi -> Prop :=
    | RemoveDupPMemberFirst: forall v set pi pi',
          removePmember v set pi pi' ->
          remove_dup_pmember (PMember v set pi) (PMember v set pi')
    | RemoveDupPMemberNext: forall v set pi pi',
          remove_dup_pmember pi pi' ->
          remove_dup_pmember (PMember v set pi) (PMember v set pi').

Theorem consequent_pi_remove_dup_pmember:
    forall pi pi' finish,
       remove_dup_pmember pi pi' ->
       consequent_pi pi' finish ->
       consequent_pi pi finish.
Admitted.

(* removes a Pmember clause from a pi which is of the form "v \in R_F(v,nil)" *)

Fixpoint remove_redundant_pmember ( p : pi ) { struct p } : pi :=
    match p with
        | PBody x => PBody x
        | (PMember v (R fields (AVar x) nil) rest) => if beq_id v x then remove_redundant_pmember rest else (PMember v (R fields (AVar x) nil) (remove_redundant_pmember rest))
        | (PMember v x rest) => (PMember v x (remove_redundant_pmember rest))
    end.

(* This is a sound principle since a vacuous R is removed; thus this axiom *)

Theorem consequent_pi_remove_redundant_pmember:
     forall pi finish,
       consequent_pi (remove_redundant_pmember pi) finish ->
       consequent_pi pi finish.
Admitted.

Theorem consequent_pi_prime_pmember:
    forall pi finish v set,
        consequent_pi pi finish ->
        var_free_pi pi false (PrimedId v) ->
        consequent_pi (PMember (PrimedId v) set pi) finish.
Admitted.

(* Predicate to check if a given "x \in \Sigma" is in a given \Pi. *)

Inductive has_pmember: id -> absHeap -> pi -> Prop :=
    | HasPmemberFirst: forall v set pi, has_pmember v set (PMember v set pi)
    | HasPmemberNext: forall v set v2 set2 pi,
        has_pmember v set pi ->
        has_pmember v set (PMember v2 set2 pi).

Inductive exists_consequent: pi -> pi -> Prop :=
    (*| ExistsConsequent1: forall pi x v l f set rest,
        andComponent (l ~ f ~> (AVar x)) (extractBody pi) rest ->
        exists_consequent (FExists v set (l ~ f ~> (AVar v))) pi*)
    | ExistsConsequent1: forall pi pi' x v l f set rest,
        (FExists v set (l ~ f ~> (AVar v))) = extractBody pi ->
        andComponent (l ~ f ~> (AVar x)) (extractBody pi') rest ->
        has_pmember x set pi' ->
        exists_consequent pi pi'
    | ExistsConsequent2: forall v set t pi pi',
        var_free_exp t false v ->
        (FExists v set ((AVar v)====t)) = extractBody pi ->
        exists_consequent pi pi'.

Theorem consequent_pi_exist_consequent: forall pi pi',
    exists_consequent pi pi' ->
    consequent_pi (replaceBody pi FTrue) pi' ->
    consequent_pi pi pi'.
Admitted.

(* Auxilliary function to check if an id is primed or not *)

Fixpoint primedID (id : id) : bool :=
    match id with
        | (PrimedId x) => true
        | _ => false
    end.

Theorem consequent_pi_eliminate_loaner: forall fields var p p' l f v rest rest',
    andComponent ((AVar l) ~ f ~> (AVar v)) (extractBody p) rest ->
    primedID v=true ->
    has_pmember l (R fields var nil) p ->
    mem_id f fields=true ->
    var_free_pi (replaceBody p rest) false v ->
    andComponent (FNot ((AVar l)====A0)) (extractBody p') rest' ->
    consequent_pi (replaceBody p rest) p' ->
    consequent_pi p p'.
Admitted.

(*Theorem consequent_pi_eliminate_loaner2: forall p p' l f v rest rest',
    andComponent (l ~ f ~> (AVar v)) (extractBody p) rest ->
    var_free_formula rest false v ->
    andComponent (FNot (l====A0)) (extractBody p') rest' ->
    consequent_pi (replaceBody p rest) p' ->
    consequent_pi p p'.
Admitted.*)

(* Conjunction c in the following is a duplicate and can soundly be eliminated. *)

Theorem consequent_pi_eliminate_and_dup:
    forall p p' rest rest2 c,
        andComponent c (extractBody p) rest ->
        andComponent c rest rest2 ->
        consequent_pi (replaceBody p rest) p' ->
        consequent_pi p p'.
Admitted.

Inductive delete_pmember: id -> absHeap -> pi -> pi -> Prop :=
    | DeletePmemberFirst: forall v set pi, delete_pmember v set (PMember v set pi) pi
    | DeletePmemberNext: forall v set v2 set2 pi pi',
        delete_pmember v set pi pi' ->
        delete_pmember v set (PMember v2 set2 pi) (PMember v2 set2 pi').

Theorem consequent_pi_eliminate_subsumed: forall p pp p' rest rest' v1 v2 s1 s2 x1 x2 f,
    has_pmember v1 s1 p ->
    delete_pmember (PrimedId v2) s2 p pp ->
    andComponent ((AVar v1) ~ f ~> (AVar (PrimedId v2))) (extractBody p) rest ->
    var_free_pi (replaceBody pp rest) false (PrimedId v2) ->
    andComponent (FAll x1 s1 (FExists x2 s2 ((AVar x1) ~ f ~> (AVar x2)))) (extractBody p') rest' ->
    consequent_pi (replaceBody p rest) p' ->
    consequent_pi p p'.
Admitted.

(* ***************************************************************************
 *
 * absExecute is the top level predicate that defines what execution of a
 * statement does to the abstract state.  It is notated
 *   \Pi | \Sigma ==c==> \Pi | \Sigma' in the paper.
 *
 ***************************************************************************)
Definition absExecute (c : com) (s : absState) (s' : absState) : Prop :=
    forall st st',
        stateMap s st ->
        c /st ~~> st' ->
        stateMap s' st'.

Notation "c1 '//' st '~~~>' st'" := (absExecute c1 st st') (at level 40).

(*
 * Here is the abstract state execution theorem for SKIP (along with its
 * proof)
 *)
Theorem execute_SKIP : forall st, SKIP // st ~~~> st.

Proof.
    unfold absExecute. intros st st0 st'. intros H. intros C. inversion C. rewrite <- H2.  apply H. Qed.

(* ***************************************************************************
 *
 * I commented out the not_present... definitions since they are not used.
 * However, I anticipate that they will be used later as I expect to need
 * to deal with variable capture issues in the near future.
 *
 *****************************************************************************)
(*Fixpoint not_present_spatial (n : id) (s : absHeap) { struct s } : Prop :=
    match s with
        | Empty => True
        | Compose a b => not_present_spatial n a /\ not_present_spatial n b
        | SpatialCellComplete f =>
          (forall id v, (f v=Some id) -> not(id=n))
        | SpatialCell f =>
          (forall id v, (f v=Some id) -> not(id=n))
        | R r v lv => not(v=n) /\ not(mem_id v lv=true)
        | Junk => True
    end.

Fixpoint not_present_exp (n : id) (e : aexp) { struct e } : Prop :=
    match e with
        | ANum n => True
        | AVar v => not(v=n)
        | APlus l r => not_present_exp n l /\ not_present_exp n r
        | AMinus l r => not_present_exp n l /\ not_present_exp n r
        | AMult l r => not_present_exp n l /\ not_present_exp n r
    end.

Fixpoint not_present_formula (n : id) (f : formula) { struct f } : Prop :=
    match f with
        | FTrue => True
        | FFalse => True
        | FEqual l r => not_present_exp n l /\ not_present_exp n r
        | FLessEqual l r => not_present_exp n l /\ not_present_exp n r
        | FAnd l r => not_present_formula n l /\ not_present_formula n r
        | FOr l r => not_present_formula n l /\ not_present_formula n r
        | FNot l => not_present_formula n l
        | FAll v f => v=n \/ not_present_formula n f
        | FExists v f => v=n \/ not_present_formula n f
        | LeadsTo s r t => not_present_exp n s /\ not_present_exp n t
        | FRef s f t => not_present_exp n s /\ not_present_exp n t
    end.

Definition not_present (n : id) (s : absState) : Prop :=
    (not_present_spatial n (fst s)) /\ (not_present_formula n (snd s)).*)

(* ***************************************************************************
 *
 * Auxiliary substitution functions to replace a variable with an
 * expression in various entities.
 *
 *****************************************************************************)
Fixpoint subst_exp (n : id) (x : aexp) (e : aexp) { struct e } : aexp :=
    match e with
        | ANum n => ANum n
        | AVar v => (if beq_id v n then x else AVar v)
        | APlus l r => APlus (subst_exp n x l) (subst_exp n x r)
        | AMinus l r => AMinus (subst_exp n x l) (subst_exp n x r)
        | AMult l r => AMult (subst_exp n x l) (subst_exp n x r)
    end.

Fixpoint subst_exp_list (n : id) (r : aexp) (vl : list aexp) { struct vl } : list aexp :=
    match vl with
        | nil => nil
        | (f::l) => (subst_exp n r f)::(subst_exp_list n r l)
    end.

Definition eqSVar (n : option id) (r : option id) : bool :=
    match n with
        | None => match r with
                      | None => true
                      | _ => false
                  end
        | Some n => match r with
                      | Some r => beq_id n r
                      | _ => false
                    end
    end.

Fixpoint subst_pair_list (n : id) (r : aexp) (p : list (id * aexp)) { struct p } : list (id * aexp) :=
    match p with
        | ((f,e)::rest) => (f,subst_exp n r e)::(subst_pair_list n r rest)
        | nil => nil
    end.

Fixpoint subst_spatial (n : id) (r : aexp) (s : absHeap) { struct s } : absHeap :=
    match s with
        | Empty => Empty
        | Compose a b => Compose (subst_spatial n r a) (subst_spatial n r b)
        | HeapImply a b => HeapImply (subst_spatial n r a) (subst_spatial n r b)
        | SpatialCell loc f => SpatialCell (subst_exp n r loc) (subst_pair_list n r f)
        | R rr v lv => R rr (subst_exp n r v) (subst_exp_list n r lv)
        | Junk => Junk
    end.

Fixpoint subst_formula (n : id) (x : aexp) (f : formula) { struct f } : formula :=
    match f with
        | FTrue => FTrue
        | FFalse => FFalse
        | FEqual l r => FEqual (subst_exp n x l) (subst_exp n x r)
        | FLessEqual l r => FLessEqual (subst_exp n x l) (subst_exp n x r)
        | FAnd l r => FAnd (subst_formula n x l) (subst_formula n x r)
        | FOr l r => FOr (subst_formula n x l) (subst_formula n x r)
        | FNot l => FNot (subst_formula n x l)
        | FAll v set f => if beq_id v n then FAll v (subst_spatial n x set) f else FAll v (subst_spatial n x set) (subst_formula n x f)
        | FExists v set f => if beq_id v n then FExists v (subst_spatial n x set) f else FExists v (subst_spatial n x set) (subst_formula n x f)
        | LeadsTo s f t => LeadsTo (subst_exp n x s) f (subst_exp n x t)
    end.

Fixpoint subst_pi (n : id) (x : aexp) (p : pi) { struct p } : pi :=
    match p with
        | PMember v set f => PMember v (subst_spatial n x set) (subst_pi n x f)
        | PBody f => PBody (subst_formula n x f)
    end.

Fixpoint remove_pmember (n : id) (p : pi) { struct p } : pi :=
    match p with
        | PMember v set f => if beq_id n v then remove_pmember n f else (PMember v set (remove_pmember n f))
        | PBody f => PBody f
    end.

(*******************************************************************************
 *
 * Here are a couple of auxiliary definitions used to define the axioms below.
 *
 *******************************************************************************)
Inductive heapHasDef : absHeap -> aexp -> (list id) -> (list aexp) -> Prop :=
     | HHDBase : forall e r el,
          heapHasDef (R r e el) e r el
     | HHDInductiveLeft: forall left right e r el,
          heapHasDef left e r el ->
          heapHasDef (Compose left right) e r el
     | HHDInductiveRight: forall left right e r el,
          heapHasDef right e r el ->
          heapHasDef (Compose left right) e r el.

Inductive heapHasRef : absHeap -> aexp -> Prop :=
     | HHRBase1 : forall e fmap,
         heapHasRef (SpatialCell e fmap) e
     | HHRInductiveLeft: forall left right e,
         heapHasRef left e ->
         heapHasRef (Compose left right) e
     | HHRInductiveRight: forall left right e,
         heapHasRef right e ->
         heapHasRef (Compose left right) e.

Inductive deleteValidFormula : absHeap -> formula -> Prop :=
    | DVFTrue : forall heap, deleteValidFormula heap FTrue
    | DVFFalse : forall heap, deleteValidFormula heap FFalse
    | DVFEqual: forall heap e1 e2, deleteValidFormula heap (FEqual e1 e2)
    | DVFLess : forall heap e1 e2, deleteValidFormula heap (FLessEqual e1 e2)
    | DVFAnd : forall heap f1 f2,
              deleteValidFormula heap f1 ->
              deleteValidFormula heap f2 ->
              deleteValidFormula heap (FAnd f1 f2)
    | DVFOr : forall heap f1 f2,
              deleteValidFormula heap f1 ->
              deleteValidFormula heap f2 ->
              deleteValidFormula heap (FOr f1 f2)
    | DVFNot : forall heap f,
              deleteValidFormula heap f ->
              deleteValidFormula heap (FNot f)
    | DVFAll  : forall heap v f set,
              deleteValidFormula heap f ->
              deleteValidFormula heap (FAll v set f)
    | DVFExists : forall heap v f set,
              deleteValidFormula heap f ->
              deleteValidFormula heap (FExists v set f)
    | DVFRef : forall heap v1 f v2,
              heapHasRef heap v1 ->
              deleteValidFormula heap (LeadsTo v1 f v2).

Inductive deleteValidFormulaPi : absHeap -> pi -> Prop :=
    | DVPMember : forall v set f h,
                  deleteValidFormulaPi h f ->
                  deleteValidFormulaPi h (PMember v set f)
    | DVPBody : forall f h,
                deleteValidFormula h f ->
                deleteValidFormulaPi h (PBody f).

(*******************************************************************************
 * Some auxiliary functions needed by calT, folding and unfolding.
 *******************************************************************************)
Fixpoint mem_heap (e : aexp) (l : list absHeap) { struct l } : bool :=
    match l with
        | ((R f v t)::r) => if beq_exp v e then true else mem_heap e r
        | (_::r) => mem_heap e r
        | Nil    => false
    end.

Fixpoint beq_exp_list (l1 : list aexp) (l2 : list aexp) { struct l1 } :=
    match l1 with
        | (a::r1) => match l2 with
                         | (b::r2) => if beq_exp a b then beq_exp_list r1 r2 else false
                         | _ => false
                     end
        | nil => match l2 with
                     | nil => true
                     | _ => false
                 end
    end.

Fixpoint beq_id_list (l1 : list id) (l2 : list id) { struct l1 } :=
    match l1 with
        | (a::r1) => match l2 with
                         | (b::r2) => if beq_id a b then beq_id_list r1 r2 else false
                         | _ => false
                     end
        | nil => match l2 with
                     | nil => true
                     | _ => false
                 end
    end.

Fixpoint findMap (p : (id * aexp)) (l : list (id * aexp)) : bool :=
    match l with
        | ((x,v)::r) => match p with (x2,v2) => if beq_id x x2 then (if beq_exp v v2 then true else findMap p r) else findMap p r end
        | nil => false
    end.

Fixpoint deleteMap (p : (id * aexp)) (l : list (id * aexp)) { struct l} : list (id * aexp) :=
    match l with
        | ((x,v)::r) => match p with (x2,v2) => if beq_id x x2 then (if beq_exp v v2 then r else deleteMap p r) else deleteMap p r end
        | nil => nil
    end.

Fixpoint sameMapFun (l1 : list (id * aexp)) (l2 : list (id * aexp)) { struct l1 } : bool :=
    match l1 with
        | ((x,v)::r) => if findMap (x,v) l2 then sameMapFun r (deleteMap (x,v) l2) else false
        | nil => match l2 with
                     | nil => true
                     | _ => false
                 end
    end.

Fixpoint beq_absHeap (h1 : absHeap) (h2 : absHeap) { struct h1 } : bool :=
    match h1 with
        | Empty => match h2 with
                       | Empty => true
                       | _ => false
                   end
        | (R fields x rest) =>
                   match h2 with
                       | (R fields2 x2 rest2) => if beq_id_list fields fields2 then
                                                     (if beq_exp x x2 then
                                                         beq_exp_list rest rest2
                                                      else false)
                                                 else
                                                     false
                       | _ => false
                   end
        | (SpatialCell id l) =>
                   match h2 with
                       | (SpatialCell id2 l2) => if beq_exp id id2 then sameMapFun l l2 else false
                       | _ => false
                   end
        | _ => false
     end.

Inductive mem_heaps: absHeap -> list absHeap -> bool -> Prop :=
   | MemHeapsFirst: forall h h1 r,
          beq_absHeap h h1=true ->
          mem_heaps h (h1::r) true
   | MemHeapsRest: forall h f r,
          mem_heaps h r true ->
          mem_heaps h (f::r) true
   | MemHeapsNil: forall h,
          mem_heaps h nil false
   | MemHeapsCons: forall h f r,
          mem_heaps h r false ->
          beq_absHeap h f=false ->
          mem_heaps h (f::r) false.

Theorem mem_heaps_single_list: forall hh f, mem_heaps hh (f::nil) true -> hh=f.
Admitted.

Fixpoint makeFoldedList (f: absHeap) { struct f } : list absHeap :=
    match f with
        | Compose l r => app (makeFoldedList l) (makeFoldedList r)
        | r => r::nil
    end.

Inductive selectHeap: absHeap -> absHeap -> absHeap -> Prop :=
    | selectHeapLeft: forall l r,
      selectHeap (Compose l r) l r
    | selectHeapRight: forall l r,
      selectHeap (Compose l r) r l
    | selectHeapLeftBranch: forall l r h rest,
      selectHeap h l r ->
      selectHeap (Compose h rest) l (Compose r rest)
    | selectHeapRightBranch: forall l r h rest,
      selectHeap h l r ->
      selectHeap (Compose rest h) l (Compose rest r).

Inductive foldedHeapFields: (list absHeap) -> (list id) -> Prop :=
    | foldedHeapFieldsEmpty: forall fields,
          foldedHeapFields nil fields
    | foldedHeapFieldsCons: forall v fields rest,
          foldedHeapFields rest fields ->
          foldedHeapFields ((R fields v nil)::rest) fields.

Inductive subtractHeap: absHeap -> absHeap -> absHeap -> Prop :=
    | subtractHeapCompose: forall h1 h2 h ht hr,
      subtractHeap h h1 ht ->
      subtractHeap ht h2 hr ->
      subtractHeap h (Compose h1 h2) hr
    | subtractHeapEmpty: forall h,
      subtractHeap h Empty h
    | subtractHeapSame: forall h,
      subtractHeap h h Empty
    | subtractHeapR: forall h fields v f h',
      selectHeap h (R fields v f) h' ->
      subtractHeap h (R fields v f) h'
    | subtractHeapCell: forall h n fmap h',
      selectHeap h (SpatialCell n fmap) h' ->
      subtractHeap h (SpatialCell n fmap) h'.

Ltac selectHeap := solve [ eapply selectHeapLeft |
                           eapply selectHeapRight |
                           eapply selectHeapLeftBranch; selectHeap |
                           eapply selectHeapRightBranch; selectHeap ].

Ltac diffHeap := solve [ eapply subtractHeapCompose; diffHeap |
                         eapply subtractHeapEmpty |
                         eapply subtractHeapSame |
                         eapply subtractHeapR; selectHeap |
                         eapply subtractHeapCell; selectHeap ].

Inductive selectHeapList: absHeap -> list absHeap -> absHeap -> Prop:=
    | selectHeapListNil: forall h,
      selectHeapList h nil h
    | selectHeapListCons: forall h f r rest r2,
      selectHeap h f r ->
      selectHeapList r rest r2 ->
      selectHeapList h (f::rest) r2.


(*******************************************************************************
 * calT is the T' definition from section 4.2 of the paper (Definition 15)
 *
 *******************************************************************************)

Inductive calT: formula -> absHeap -> absHeap -> absHeap -> (id -> option absHeap) -> (id -> option absHeap) -> formula -> Prop :=
    | calTFTrue : forall s1 s2 s3 m1 m2,
                  calT FTrue s1 s2 s3 m1 m2 FTrue
    | calTFFalse : forall s1 s2 s3 m1 m2,
                   calT FTrue s1 s2 s3 m1 m2 FFalse
    | calTFAnd : forall f1 f1r f2 f2r s1 s2 s3 m1 m2,
                 calT f1 s1 s2 s3 m1 m2 f1r ->
                 calT f2 s1 s2 s3 m1 m2 f2r ->
                 calT (FAnd f1 f2) s1 s2 s3 m1 m2 (FAnd f1r f2r)
    | calTFOr : forall f1 f1r f2 f2r s1 s2 s3 m1 m2,
                calT f1 s1 s2 s3 m1 m2 f1r ->
                calT f2 s1 s2 s3 m1 m2 f2r ->
                calT (FOr f1 f2) s1 s2 s3 m1 m2 (FOr f1r f2r)
    | calTFNot : forall f fr s1 s2 s3 m1 m2,
                 calT f s1 s2 s3 m1 m2 fr ->
                 calT (FNot f) s1 s2 s3 m1 m2 (FNot fr)
    | calTFAll : forall v set f fr s1 s2 s3 m1 m2,
                 calT f s1 s2 s3 (fun x => if beq_id x v then Some set else m1 x) (fun x => if beq_id x v then None else m2 x) fr ->
                 calT (FAll v set f) s1 s2 s3 m1 m2 (FAll v set fr)
    | calTFExists : forall v set f fr s1 s2 s3 m1 m2,
                    calT f s1 s2 s3 (fun x => if beq_id x v then Some set else m1 x) (fun x => if beq_id x v then None else m2 x) fr ->
                    calT (FExists v set f) s1 s2 s3 m1 m2 (FExists v set fr)
    | calTFEqual : forall l r s1 s2 s3 m1 m2,
                 calT (FEqual l r) s1 s2 s3 m1 m2 (FEqual l r)
    | calTFLessEqual : forall l r s1 s2 s3 m1 m2,
                       calT (FLessEqual l r) s1 s2 s3 m1 m2 (FLessEqual l r)
    | calTLeadsTo1 : forall s f d s1 s2 s3 m1 m2 ss fmap e,
                     (m1 s)=Some s2 ->
                     (m2 s)=Some s3 ->
                     (SpatialCell ss fmap)=s2 ->
                     (fieldValue f fmap)=Some e ->
                     calT (LeadsTo (AVar s) f (AVar d)) s1 s2 s3 m1 m2 (FEqual (AVar d) e)
    | calTLeadsTo2 : forall s f d s1 s2 s3 m1 m2 rr ss,
                     Some ss = (m1 s) ->
                     selectHeap s1 ss rr ->
                     calT (LeadsTo (AVar s) f (AVar d)) s1 s2 s3 m1 m2 (LeadsTo (AVar s) f (AVar d))
    | calTLeadsTo3 : forall s f d s1 s2 s3 m1 m2 ss rr,
                     Some ss = (m2 s) ->
                     selectHeap s1 ss rr ->
                     calT (LeadsTo (AVar s) f (AVar d)) s1 s2 s3 m1 m2 (LeadsTo (AVar s) f (AVar d))
    | calTLeadsTo4 : forall s f d s1 s2 s3 m1 m2 ss fmap ss' fmap',
                     (m1 s)=Some s2 ->
                     (m2 s)=Some s3 ->
                     (SpatialCell ss fmap)=s2 ->
                     (SpatialCell ss' fmap')=s3 ->
                     (fieldValue f fmap)=(fieldValue f fmap') ->
                     calT (LeadsTo (AVar s) f (AVar d)) s1 s2 s3 m1 m2 (LeadsTo (AVar s) f (AVar d)).

Fixpoint addDefaultMap (m: id -> option absHeap) (h: absHeap) { struct h } : id -> option absHeap :=
         match h with
               | (SpatialCell (AVar v) map) => (fun x => if beq_id v x then Some (SpatialCell (AVar v) map) else m x)
               | (R fields (AVar v) nil) => (fun x => if beq_id v x then Some (R fields (AVar v) nil) else m x)
               | (Compose l r) => addDefaultMap (addDefaultMap m l) r
               | _ => m
         end.

Inductive calTPi: pi -> absHeap -> absHeap -> absHeap -> (id -> option absHeap) -> (id -> option absHeap) -> pi -> Prop :=
    | calTPiPMember : forall f fr s1 s2 s3 m1 m2 v set,
                      calTPi f s1 s2 s3 (fun x => if beq_id x v then Some set else m1 x) (fun x => if beq_id x v then Some set else m2 x) fr ->
                      calTPi (PMember v set f) s1 s2 s3 m1 m2 (PMember v set fr)
    | calTPiPBody: forall f fr s1 s2 s3 m1 m2,
                   calT f s1 s2 s3 (addDefaultMap (addDefaultMap m1 s1) s2) (addDefaultMap (addDefaultMap m2 s1) s3) fr ->
                   calTPi (PBody f) s1 s2 s3 m1 m2 (PBody fr).

Ltac calT := (eapply calTPiPMember; try calT) ||
             (eapply calTPiPBody; try calT) ||
             (*(eapply calTLeadsTo1; try calT) ||
             (eapply calTLeadsTo2; try calT) ||
             (eapply calTLeadsTo3; try calT) ||*)
             (eapply calTFEqual; try calT) ||
             (eapply calTFLessEqual; try calT) ||
             (eapply calTFAll; try calT) ||
             (eapply calTFExists; try calT) ||
             (eapply calTFNot; try calT) ||
             (eapply calTFOr; try calT) ||
             (eapply calTFAnd; try calT) ||
             (eapply calTFTrue; try calT) ||
             (eapply calTFFalse; try calT) ||
             simpl ||
             auto.

(*******************************************************************************
 * fold
 *
 * Here is the rule for folding states together. *** Not yet described well in
 * the paper ***
 *
 *******************************************************************************)
Inductive foldFormula: formula -> formula -> pi ->(list absHeap) -> absHeap -> Prop :=
    | FoldFormulaTrue : forall l e f,
                        foldFormula FTrue FTrue f l e
    | FoldFormulaFalse : forall l e f,
                        foldFormula FFalse FFalse f l e
    | FoldFormulaAnd : forall l r l' r' f ff e,
                 foldFormula l l' ff f e ->
                 foldFormula r r' ff f e ->
                 foldFormula (FAnd l r) (FAnd l' r') ff f e
    | FoldFormulaOr : forall l r l' r' f ff e,
                 foldFormula l l' ff f e ->
                 foldFormula r r' ff f e ->
                 foldFormula (FOr l r) (FOr l' r') ff f e
    | FoldFormulaNot : forall l l' ff f e,
                 foldFormula l l' ff f e ->
                 foldFormula (FNot l) (FNot l') ff f e
    | FoldFormulaAll : forall v set f f' ff l e,
                 mem_heaps set l false ->
                 foldFormula f f' ff l e ->
                 foldFormula (FAll v set f) (FAll v set f') ff l e
    | FoldFormulaExists : forall v set f f' ff l e,
                 mem_heaps set l false ->
                 foldFormula f f' ff l e ->
                 foldFormula (FExists v set f) (FExists v set f') ff l e
    | FoldFormulaAll2 : forall var fields v set f f' ff l e,
                 mem_heaps set l true ->
                 e = (R fields var nil) ->
                 consequent_pi (PBody (subst_formula v var f)) ff ->
                 (forall r, (mem_heaps r l true) -> consequent_pi (PBody (FAll v r f)) ff) ->
                 foldFormula f f' ff l e ->
                 foldFormula (FAll v set f) (FAll v e f') ff l e
    | FoldFormulaExists2 : forall var fields v set f f' ff l e,
                 mem_heaps set l true ->
                 e = (R fields var nil) ->
                 consequent_pi (PBody (subst_formula v var f)) ff ->
                 (forall r, (mem_heaps r l true) -> consequent_pi (PBody (FAll v r f)) ff) ->
                 consequent_pi (PBody (FExists v e f)) ff ->
                 foldFormula f f' ff l e ->
                 foldFormula (FExists v set f) (FExists v e f') ff l e
    | FoldFormulaEqual : forall l r ff f e,
                 foldFormula (FEqual l r) (FEqual l r) ff f e
    | FoldFormulaLessEqual : forall l r ff f e,
                 foldFormula (FLessEqual l r) (FLessEqual l r) ff f e
    | FoldFormulaLeadsTo : forall s f d fff e ff,
                 foldFormula (LeadsTo s f d) (LeadsTo s f d) ff fff e.

Fixpoint addDefaultPmembers (p: pi) (h: absHeap) { struct h } : pi :=
         match h with
               | (R fields (AVar v) nil) => (PMember v (R fields (AVar v) nil) p)
               | (Compose l r) => addDefaultPmembers (addDefaultPmembers p l) r
               | _ => p
         end.

Inductive foldPi: pi -> pi -> pi -> (list absHeap) -> absHeap -> Prop :=
    | FoldPiPmember: forall v r p p' l e pp,
      mem_heap (AVar v) l=false ->
      foldPi p p' pp l e ->
      foldPi (PMember v r p) (PMember v r p') pp l e
    | FoldPiPbody: forall f f' pp l e,
      foldFormula f f' pp l e ->
      foldPi (PBody f) (PBody f') pp l e.

Ltac reduceFold := (apply FoldFormulaTrue; try reduceFold) ||
                      (apply FoldFormulaFalse; try reduceFold) ||
                      (apply FoldFormulaAnd; try reduceFold) ||
                      (apply FoldFormulaOr; try reduceFold) ||
                      (apply FoldFormulaNot; try reduceFold) ||
                      (*(apply FoldFormulaAll2; try reduceFold) ||
                      (apply FoldFormulaExists2; try reduceFold) ||
                      (apply FoldFormulaAll; try reduceFold) ||
                      (apply FoldFormulaExists; try reduceFold) ||*)
                      (apply FoldFormulaEqual; try reduceFold) ||
                      (apply FoldFormulaLessEqual; try reduceFold) ||
                      (apply FoldFormulaLeadsTo; try reduceFold) ||
                      (apply FoldPiPmember; try reduceFold) ||
                      (apply FoldPiPbody; try reduceFold) ||
                      (apply MemHeapsRest; try reduceFold) ||
                      (apply MemHeapsNil; try reduceFold) ||
                      (apply MemHeapsCons; try reduceFold) ||
                      (apply MemHeapsFirst; try reduceFold) ||
                      (repeat auto).

Inductive foldHeap: absState -> absState -> Prop :=
    | foldRule: forall cell fmap fields v heaps foldedHeaps rh remainingHeaps pi pi',
          selectHeap heaps cell rh ->
          selectHeapList rh foldedHeaps remainingHeaps ->
          cell = SpatialCell v fmap ->
          (forall e, mem_heap e foldedHeaps=true -> exists id, fieldValue id fmap = Some e) ->
          foldedHeapFields foldedHeaps fields ->
          foldPi pi pi' pi foldedHeaps (R fields v nil) ->
          foldHeap (heaps,pi)
                   ((Compose (R fields v nil) remainingHeaps),pi').

(*******************************************************************************
 * mergeStates
 *
 * This is the predicate used to merge two states. *** Not yet described well in
 * the paper ***
 *
 *******************************************************************************)

Inductive add_pmember: absHeap -> pi -> pi -> Prop :=
    | AddPMember: forall fields var pi heap rest,
        selectHeap heap (R fields (AVar var) nil) rest ->
        add_pmember heap pi (PMember var (R fields (AVar var) nil) pi)
    | AddPMemberDropLoner: forall body rest rest2 x y f field pi heap,
                                  body = extractBody pi ->
                                  andComponent (f ~ field ~> (AVar x)) body rest ->
                                  andComponent (f ~ field ~> y) rest rest2 ->
                                  var_free_spatial heap true x ->
                                  var_free_pi pi true x ->
                                  add_pmember heap pi (replaceBody pi rest)
    | AddPMemberNone: forall heap pi, add_pmember heap pi pi.

Inductive mergeStates: absState -> absState -> absState -> formula -> Prop :=
    | mergeStatesSame : forall spatial2 pi1 pi2 spatial pi f pi1' pi2',
          subtractHeap spatial2 spatial Empty ->
          add_pmember spatial pi1 pi1' ->
          add_pmember spatial pi2 pi2' ->
          consequent_pi pi1' (addFormula pi f) ->
          consequent_pi pi2' (addFormula pi (FNot f)) ->
          mergeStates (spatial,pi1) (spatial2,pi2) (spatial,pi) f
    | mergeStatesFoldLeft : forall right final spatial spatial' pi pi' f,
          foldHeap (spatial,pi) (spatial',pi') ->
          mergeStates (spatial',pi') right final f ->
          mergeStates (spatial,pi) right final f
    | mergeStatesFoldRight : forall left final spatial spatial' pi pi' f,
          foldHeap (spatial,pi) (spatial',pi') ->
          mergeStates left (spatial',pi') final f ->
          mergeStates left (spatial,pi) final f.

(*******************************************************************************
 * unfold rule
 * This is the arr1 rule and unf function from section 4.2 (figure 3)
 *******************************************************************************)

Inductive buildHeaps: absHeap -> list id -> list id -> absHeap -> pi -> list id -> list (id * aexp) -> Prop :=
    | buildHeapsNil : forall h pi allFields fn fields,
                      buildHeaps h nil fields h pi allFields fn
    | buildHeapsCons : forall allHeaps restHeaps f v vlist pi allFields fn fields fn2,
                       buildHeaps allHeaps fields allFields restHeaps pi vlist fn ->
                       var_free_pi pi true v ->
                       var_free_spatial allHeaps true v ->
                       fn2 = (f,(AVar v))::fn ->
                       buildHeaps (Compose allHeaps (R allFields (!v) nil)) (f::fields) allFields restHeaps pi (v::vlist) fn2.

Inductive buildBody: formula -> formula -> id -> (list id) -> formula -> Prop :=
    | buildBodyNil: forall body b v,
                    buildBody body b v nil b
    | buildBodyCons: forall body b v f rest body',
                     buildBody body b v rest body' ->
                     buildBody body b v (f::rest) (FAnd (subst_formula v (AVar f) body) body').

Inductive buildVarList: formula -> (list id) -> (list id) -> Prop :=
    | buildVarListNil : forall f,
                        buildVarList f nil nil
    | buildVarListCons : forall formula fields f v vars,
                         buildVarList (FAnd formula (FEqual (!v) (!v))) fields vars ->
                         var_free_formula formula true v ->
                         buildVarList formula (f::fields) (v::vars).

Inductive buildNewPi: (list id) -> (list id) -> id -> pi -> pi -> Prop :=
    | buildNewPiNil: forall p fields v,
                     buildNewPi nil fields v p p
    | buildNewPiCons: forall p np v rest var fields,
                      buildNewPi rest fields var p np ->
                      buildNewPi (v::rest) fields var p (PMember v (R fields (!var) nil) np).

Inductive buildAllList: formula -> (list id) -> id -> id -> (list id) -> formula -> Prop :=
    | buildAllListNil: forall f fields v var,
                       buildAllList f fields v var nil (subst_formula v (AVar var) f)
    | buildAllListCons: forall f res fields v var vfirst vrest,
                        buildAllList f fields v var vrest res ->
                        buildAllList f fields v var (vfirst::vrest) (FAnd (FAll v (R fields (AVar vfirst) nil) f) res).

Inductive buildExistsList: formula -> (list id) -> id -> id -> (list id) -> formula -> Prop :=
    | buildExistsListNil: forall f fields v var,
                       buildExistsList f fields v var nil (subst_formula v (AVar var) f)
    | buildExistsListCons: forall f res fields v var vfirst vrest,
                        buildExistsList f fields v var vrest res ->
                        buildExistsList f fields v var (vfirst::vrest) (FOr (FExists v (R fields (AVar vfirst) nil) f) res).

Inductive unf_formula: formula -> formula -> absHeap -> (list id) -> absHeap -> Prop :=
    | unfTrue : forall h l r,
                unf_formula FTrue FTrue h l r
    | unfFalse : forall h l r,
                 unf_formula FFalse FFalse h l r
    | unfAnd : forall l r l' r' heap vars rest,
                 unf_formula l l' heap vars rest ->
                 unf_formula r r' heap vars rest ->
                 unf_formula (FAnd l r) (FAnd l' r') heap vars rest
    | unfOr : forall l r l' r' heap vars rest,
                 unf_formula l l' heap vars rest ->
                 unf_formula r r' heap vars rest ->
                 unf_formula (FOr l r) (FOr l' r') heap vars rest
    | unfNot : forall l l' heap vars rest,
                 unf_formula l l' heap vars rest ->
                 unf_formula (FNot l) (FNot l') heap vars rest
    | unfAll : forall v set f f' heap vars fields rest v2 fields2 v3,
                 set = R fields (AVar v2) nil ->
                 heap = R fields2 (AVar v3) nil ->
                 beq_id v3 v2=false ->
                 unf_formula f f' heap vars rest ->
                 unf_formula (FAll v set f) (FAll v set f') heap vars rest
    | unfExists : forall v set f f' heap vars rest v2 fields fields2 v3,
                 set = R fields (AVar v2) nil ->
                 heap = R fields2 (AVar v3) nil ->
                 beq_id v3 v2=false ->
                 unf_formula f f' heap vars rest ->
                 unf_formula (FExists v set f) (FExists v set f') heap vars rest
    | unfAll2 : forall v set f f' heap vars fields rest v2 res fields2 v3,
                 set = R fields (AVar v2) nil ->
                 heap = R fields2 (AVar v3) nil ->
                 beq_id v3 v2=true ->
                 unf_formula f f' heap vars rest ->
                 buildAllList f fields v v2 vars res ->
                 unf_formula (FAll v set f) res heap vars rest
    | unfExists2 : forall v set f f' heap vars fields rest v2 res fields2 v3,
                 set = R fields (AVar v2) nil ->
                 heap = R fields2 (AVar v3) nil ->
                 beq_id v3 v2=true ->
                 unf_formula f f' heap vars rest ->
                 buildExistsList f fields v v2 vars res ->
                 unf_formula (FExists v set f) res heap vars rest
    | unfEqual : forall l r heap vars rest,
                 unf_formula (FEqual l r) (FEqual l r) heap vars rest
    | unfLessEqual : forall l r heap vars rest,
                 unf_formula (FLessEqual l r) (FLessEqual l r) heap vars rest
    | unfLeadsTo : forall s f d heap vars rest,
                 unf_formula (LeadsTo s f d) (LeadsTo s f d) heap vars rest.

Inductive unf_pi: pi -> pi -> absHeap -> (list id) -> absHeap -> Prop :=
    | unfPMember1 : forall v f f' fields fields1 var var1 vars restHeaps,
                    unf_pi f f' (R fields (AVar var) nil) vars restHeaps ->
                    beq_id var var1=false ->
                    unf_pi (PMember v (R fields1 (AVar var1) nil) f) (PMember v (R fields1 (AVar var1) nil) f') (R fields (AVar var) nil) vars restHeaps
    | unfPMember2 : forall f f' fields var vars restHeaps body v newPi fullBody,
                    unf_pi f f' (R fields (AVar var) nil) vars restHeaps ->
                    body = extractBody f ->
                    buildVarList body fields vars ->
                    buildBody body (subst_formula v (AVar var) body) v vars fullBody ->
                    buildNewPi vars fields v (replaceBody f fullBody) newPi ->
                    unf_pi (PMember v (R fields (AVar var) nil) f) newPi (R fields (AVar var) nil) vars restHeaps
    | unfPBody : forall f f' r vars restHeaps,
                 unf_formula f f' r vars restHeaps ->
                 unf_pi (PBody f) (PBody f') r vars restHeaps.

Ltac unf := (apply unfPMember1; try unf) ||
            (apply unfPMember2; try unf) ||
            (apply unfPBody; try unf) ||
            (apply unfTrue; try unf) ||
            (apply unfFalse; try unf) ||
            (apply unfAnd; try unf) ||
            (apply unfOr; try unf) ||
            (apply unfNot; try unf) ||
            (apply unfAll; try unf) ||
            (apply unfExists; try unf) ||
            (apply unfEqual; try unf) ||
            (apply unfLessEqual; try unf) ||
            (apply unfLeadsTo; try unf) ||
            auto.

Inductive unfold_heap: absState -> absState -> Prop :=
    unfoldRule: forall spatial pi spatial' pi' restHeap fields var allHeaps vars fvar,
                selectHeap spatial (R fields var nil) restHeap ->
                buildHeaps allHeaps fields fields restHeap pi vars fvar ->
                unf_pi pi pi' (R fields var nil) vars restHeap ->
                spatial' = Compose (SpatialCell var fvar) allHeaps ->
                unfold_heap (spatial,pi) (spatial',pi').

(*******************************************************************************
 *
 * Finally, here are the axioms that deal with the various state modifying
 * statements.  We will not need axioms for the sequence, if or while constructs
 * as they are already taken care of in Hoare.v.  The axioms here correspond
 * to section to the as.. rules from section 4.2 of the paper.
 *
 *******************************************************************************)
Theorem execute_ASSIGN : forall spatial formula spatial' formula' v e,
     spatial' = subst_spatial v e spatial ->
     formula' = subst_pi v e (remove_pmember v formula) ->
     (v ::= e) // (spatial',formula') ~~~> (spatial,formula).
Admitted.

Theorem execute_DELETE : forall spatial formula spatial' e fmap,
     (*deleteValidFormulaPi spatial formula ->*)
     spatial' = (Compose spatial (SpatialCell e fmap)) ->
     (DELETE e) // (spatial',formula) ~~~> (spatial,formula).
Admitted.

Theorem execute_NEW : forall spatial formula spatial' nv v formula',
     var_free_spatial spatial false nv ->
     var_free_pi formula false nv ->
     spatial' = (HeapImply (SpatialCell (AVar nv) nil) (subst_spatial v (AVar nv) spatial)) ->
     formula' = subst_pi v (AVar nv) formula ->
     (NEW v) // (spatial',formula') ~~~> ((Compose (SpatialCell (AVar nv) nil) spatial),formula).
Admitted.

Theorem execute_LOAD : forall v nv spatial formula spatial' formula' f e,
     var_free_spatial spatial false nv ->
     var_free_pi formula false nv ->
     spatial' = subst_spatial v (AVar nv) spatial ->
     formula' = (addFormula (subst_pi v (AVar nv) (remove_pmember v formula)) (LeadsTo e f (AVar nv))) ->
     (CLoad v e f) // (spatial',formula') ~~~> (spatial,formula).
Admitted.

Theorem execute_STORE : forall spatial formula spatial' formula' v fmap fmap' e f rest,
     spatial = Compose rest (SpatialCell (AVar v) fmap) ->
     sameMap fmap ((v,e)::fmap') ->
     calTPi formula spatial (SpatialCell (AVar v) fmap) (SpatialCell (AVar v) fmap') (fun x => None) (fun x => None) formula' ->
     spatial' = Compose rest (SpatialCell (AVar v) fmap') ->
     (CStore (AVar v) f e) // (spatial',formula') ~~~> (spatial,formula).
Admitted.