(* Hoare.v: Hoare Logic *)
(* Based on work from http://www.seas.upenn.edu/~cis500/current/index.html *)

(* This file implements a Hoare logic for the abstract states in
   AbsHeap.v.   *)

Require Export ImpHeap.
Require Export AbsHeap.

(* Assertions over concrete heap states *)

Definition Assertion := state -> Prop.

(* Hoare triples over concrete heap states *)

Definition hoare_triple (P:Assertion) (c:com) (Q:Assertion) : Prop :=
  forall st st', 
       c / st ~~> st' 
    -> P st 
    -> Q st'.

(* The next two definitions create Hoare notation for the
   abstract states from AbsHeap.v, over which we will be doing our proofs *)

Definition absStateAssertion (st : absState) : Assertion :=
    (fun s => stateMap st s).

Notation "{{ P }} c {{ Q }}" := (hoare_triple (absStateAssertion P) c (absStateAssertion Q)) (at level 90).


(*
 * This theorem creates a Hoare logic front end for the execute_ASSIGN
 * theorem in AbsHeap.v
 *)

Theorem hoare_abs_asgn : forall spatial' spatial formula' formula v e formula2,
    subst_pi v e (remove_pmember v formula') = formula2 ->
    subst_spatial v e spatial' = spatial ->
    consequent_pi formula2 formula ->
    {{(spatial,formula)}}(v ::= e){{(spatial',formula')}}.
Admitted.

(*
 * This theorem creates a Hoare logic front end for the execute_DELETE, ...
 * theorems in AbsHeap.v.  They are based on as1-as5 from section 4.2 of the paper
 *)
Theorem hoare_abs_delete : forall spatial' spatial formula fmap e,
     spatial' = (Compose spatial (SpatialCell e fmap)) ->
     {{(spatial',formula)}}DELETE e{{(spatial,formula)}}.

Admitted.

Theorem hoare_abs_new : forall spatial formula v rest,
     selectHeap spatial (SpatialCell (AVar v) nil) rest ->
     var_free_spatial rest false v ->
     var_free_pi formula false v ->
     {{(rest,formula)}}(NEW v){{(spatial,formula)}}.
Admitted.

Theorem hoare_abs_load : forall v nv spatial formula spatial' formula' f e formula2,
     (*var_free_spatial spatial false nv ->
     var_free_pi formula false nv ->*)
     spatial' = subst_spatial v nv spatial ->
     formula' = (addFormula (subst_pi v nv (remove_pmember v formula)) (LeadsTo e f nv)) ->
     consequent_pi formula' formula2 ->
     {{(spatial',formula2)}}(CLoad v e f){{(spatial,formula)}}.
Admitted.

Theorem hoare_abs_store : forall spatial formula spatial' formula' v fmap fmap' e f rest oldValue,
     selectHeap spatial (SpatialCell (AVar v) fmap) rest ->
     fieldValue f fmap = Some e ->
     fmap' = replaceValue fmap f oldValue ->
     calTPi formula rest (SpatialCell (AVar v) fmap) (SpatialCell (AVar v) fmap') (fun x => None) (fun x => None) formula' ->
     spatial' = Compose rest (SpatialCell (AVar v) fmap') ->
     {{(spatial',formula')}}(CStore (AVar v) f e){{(spatial,formula)}}.
Admitted.


(* Rules of consequence *)


(*
 * The next two theorems are the standard hoare consequence theorems and
 * allow use of the equivalent_spatial and equivalent_formula theorems in
 * AbsHeap.v
 *)

Theorem hoare_consequence_abs_equivalent_pre :
    forall spatial formula spatial' formula' Q c,
        {{(spatial',formula')}}c{{Q}} ->
        equivalent_spatial spatial spatial' ->
        equivalent_pi formula formula' ->
        {{(spatial,formula)}}c{{Q}}.
Admitted.

Theorem hoare_consequence_abs_pre :
    forall spatial formula formula' Q c,
        consequent_pi formula' formula ->
        {{(spatial,formula')}}c{{Q}} ->
        {{(spatial,formula)}}c{{Q}}.
Admitted.

(* Recasting of the unfold rule from rule arr1 *)

Theorem hoare_unfold_post : forall P Q Q' c,
        unfold_heap Q Q' ->
        {{P}}c{{Q'}} ->
        {{P}}c{{Q}}.
Admitted.


(* Proof rule for SKIP *)

Theorem hoare_skip_abs : forall P,
     {{P}} SKIP {{P}}.
Proof.
  unfold hoare_triple. intros P st st' H HP. inversion H. subst.
  assumption.  Qed.


(* Proof rule for sequencing ; *)


Theorem hoare_seq_abs : forall P Q R c1 c2,
     {{Q}} c2 {{R}}
  -> {{P}} c1 {{Q}}
  -> {{P}} c1;c2 {{R}}.
Proof.
  unfold hoare_triple.
  intros P Q R c1 c2 H1 H2 st st' H12 Pre.
  inversion H12; subst.
  apply (H1 st'0 st'); try assumption.  
  apply (H2 st st'0); try assumption. Qed.

(* Some preliminary definitions are needed for the conditional rule *)


Definition bassn b : Assertion :=
  fun st => beval st b = true.

(** A couple of facts about [bassn] will be useful in proving the
    validity of the proof rule for conditionals. *)

Lemma bexp_eval_true : forall b st,
  beval st b = true -> (bassn b) st.
Proof.
  intros b st Hbe.
  unfold bassn. assumption.  Qed.

Lemma bexp_eval_false : forall b st,
  beval st b = false -> ~ ((bassn b) st).
Proof.
  intros b st Hbe contra.
  unfold bassn in contra.
  rewrite -> contra in Hbe. inversion Hbe.  Qed.

Fixpoint convert_to_formula (b : bexp) { struct b } : formula :=
    match b with
    | BTrue => FTrue
    | BFalse => FFalse
    | BEq a b => (FEqual a b)
    | BLe a b => (FLessEqual a b)
    | BNot x => (FNot (convert_to_formula x))
    | BAnd a b => (FAnd (convert_to_formula a) (convert_to_formula b))
    | BOr a b => (FOr (convert_to_formula a) (convert_to_formula b))
    end.

(*
 * Proof rule for if
 *)

Theorem hoare_if_abs : forall spatial1' spatial2' formula1'' formula2'' formula1' formula2' spatial' formula' spatial formula b c1 c2 f,
  f = convert_to_formula b ->
  {{(spatial1',formula1')}} c1 {{(spatial,formula)}} ->
  {{(spatial2',formula2')}} c2 {{(spatial,formula)}} ->
  consequent_pi formula1' formula1'' ->
  consequent_pi formula2' formula2'' ->
  mergeStates (spatial1',formula1'') (spatial2',formula2'') (spatial',formula') f ->
  {{(spatial',formula')}} (CIf b c1 c2) {{(spatial,formula)}}.
Admitted.


(* 
 * Proof rule for while 
 *)

Lemma hoare_while_abs : forall spatial formula c b f formula' invariant_pi invariant_spatial,
  f = convert_to_formula b ->
  {{(invariant_spatial,addFormula invariant_pi f)}} c {{(invariant_spatial,invariant_pi)}} ->
  consequent_pi formula' (addFormula formula (FNot f)) ->
  consequent_spatial invariant_spatial spatial ->
  consequent_pi invariant_pi formula ->
  {{(spatial,formula)}} (WHILE b DO c LOOP) {{(spatial,formula')}}.
Admitted.


(* **************************************************************************
 *
 * Here is the example from the paper.  We start with some definitions for
 * the variables in the program.
 *
 ***************************************************************************)

Definition P : id := Id 0.
Definition RR : id := Id 1.
Definition I : id := Id 2.
Definition N : id := Id 3.
Definition T : id := Id 4.
Definition Tmp_l : id := Id 5.
Definition Tmp_r : id := Id 6.

(*
 * These definitions are for the fields of both the list and the tree
 * struct.  Note that the two struct types are not distinguished, the
 * programs have no types.
 *)
Definition F_l : id := Id 0.
Definition F_r : id := Id 1.
Definition F_n : id := Id 0.
Definition F_p : id := Id 1.

(*
 * And here are a couple of extra variables to be used in the quantifiers.
 *)
Definition vn : id := Id 7.
Definition vm : id := Id 8.

(*
 * Here is the abstract state condition at the beginning of the program.
 * The \Sigma portion of the states that the variable RR points to a well
 * formed tree.  There are no predicates in the \Pi portion.
 *)
Definition precondition : absState := (
    (R (F_l::F_r::nil) (!RR) nil),
    (PBody FTrue)).

(*
 * This assertion is a little tricky.  It is the invariant for the main
 * loop of our example. We still have the assertion that RR points to a
 * well formed tree.  We also have two additional data structures in the
 * heap.  Both are linked lists.  The first one is pointed to by the
 * variable I and the second one is pointed to by the variable P.
 *
 * The \Pi portion in the assertion below states that the variable T
 * is either nil or it points to a node inside the tree whose top is
 * stored in RR.
 * There is a second condition in \Pi which states that all of the
 * cells in the list headed by P, the F_p field points to a node in the
 * tree RR.  This is the condition that uses the quantifiers in the \Pi
 * portion.
 *)
Definition afterInitAssigns : absState := (
    ( (R (F_l::F_r::nil) (!RR) nil) **
      (R (F_n::nil) (!I) nil) **
      (R (F_n::nil) (!P) nil)),
    (
        (PMember T (R (F_l::F_r::nil) (!RR) nil) (PBody
            (FAnd
                (FAll vn (R (F_n::nil) (!P) nil)
                    (FExists vm (R (F_l::F_r::nil) (!RR) nil)
                        (LeadsTo (!vn) (F_p) (!vm))
                    )
                )
                (FAll vn (R (F_n::nil) (!I) nil)
                    (FExists vm (R (F_l::F_r::nil) (!RR) nil)
                        (LeadsTo (!vn) (F_p) (!vm))
                    )
                )
            )
        ))
    )).

(*
 * Here are the first three lines of code from the example
 *)
Definition initCode : com :=
    I ::= A0;
    T ::= !RR;
    P ::= A0.

(*
 * ...and here is the proof that after these lines are executed, that
 * we indeed have the state described by "afterInitAssigns".  Actually,
 * the state afterInitAssigns is a generalization of the actual state.
 * Most of the proof involves simplifying the clauses after nil is filled
 * in for the variables I and P and after T is replaced with RR.
 *)
Theorem initialization : {{precondition}}initCode{{afterInitAssigns}}.

Proof.

    (* Start by unfolding stuff *)
    unfold initCode. eapply hoare_consequence_abs_equivalent_pre.

    (* Now let's apply hoare_seq_abs twice to break up the sequences *)
    eapply hoare_seq_abs. eapply hoare_seq_abs.

    (* Working backwards, we back-propagate the state over the third
       assign. *)
    eapply hoare_abs_asgn. simpl. reflexivity. simpl. reflexivity.
    eapply consequent_pi_reflexive.

    (* Now we back propagate over the second assign *)
    eapply hoare_abs_asgn. simpl. reflexivity. simpl. reflexivity.
    eapply consequent_pi_reflexive. 

    (* And finally, we back-propagate over the first assignment. *)
    eapply hoare_abs_asgn. simpl. reflexivity. simpl. reflexivity.
    eapply consequent_pi_reflexive.

    simpl. eauto 50. simpl.

    (* Now, we have the job of proving the back-propagated state equivalent
       to our initial state.  This is done by applying a long sequence of the
       equivalent_spatial... theorems *)
    (*eapply equivalent_spatial_compose2.
    apply equivalent_empty_R.
    apply equivalent_spatial_reflexive.
    eapply equivalent_empty_compose2.
    eapply equivalent_spatial_compose2.
    apply equivalent_empty_R.
    apply equivalent_spatial_reflexive.
    apply equivalent_empty_compose2.
    apply equivalent_spatial_reflexive.

    simpl.*)

    (* Now we need to prove the formula portions of the two states.  This
       is done by applying a long sequence of the equivalent_formula...
       theorems. *)

    eapply equivalent_pi_pbody.
    eapply equivalent_formula_and1.
    eapply equivalent_formula_all_set.
    eapply equivalent_empty_R.
    eapply equivalent_spatial_reflexive.
    eapply equivalent_formula_all_true_empty.
    eapply equivalent_formula_reflexive.
    eapply equivalent_formula_and2.
    eapply equivalent_formula_all_set.
    eapply equivalent_empty_R.
    eapply equivalent_spatial_reflexive.
    eapply equivalent_formula_all_true_empty.
    eapply equivalent_formula_reflexive.
    eapply equivalent_formula_and_simp.
    simpl. eapply equivalent_formula_reflexive.
    eapply equivalent_pi_reflexive.

Qed.

(*
 * Now we define the abstract state that should exist after the while of
 * the main body of the program is executed.
 *
 * This is exactly the same as the state afterInitAssigns except that the
 * condition T=0 and I=0 have been added.
 *)
Definition afterWhile : absState := (
    ( (R (F_l::F_r::nil) (!RR) nil) **
      (R (F_n::nil) (!I) nil) **
      (R (F_n::nil) (!P) nil)),
    (
        (PMember T (R (F_l::F_r::nil) (!RR) nil)
            (PBody
                (FAnd
                    (FAll vn (R (F_n::nil) (!P) nil)
                        (FExists vm (R (F_l::F_r::nil) (!RR) nil)
                            (LeadsTo (!vn) (F_p) (!vm))
                        )
                    )
                    (FEqual (!T) A0)
                )
            )
        )
    )).

Definition loopSpatial :=
      (R (F_l::F_r::nil) (!RR) nil) **
      (R (F_n::nil) (!I) nil) **
      (R (F_n::nil) (!P) nil).

Definition loopPi :=
        (PMember T (R (F_l::F_r::nil) (!RR) nil)
            (PBody
                (FAnd
                    (FAll vn (R (F_n::nil) (!P) nil)
                        (FExists vm (R (F_l::F_r::nil) (!RR) nil)
                            (LeadsTo (!vn) (F_p) (!vm))
                        )
                    )
                    (FAll vn (R (F_n::nil) (!I) nil)
                        (FExists vm (R (F_l::F_r::nil) (!RR) nil)
                            (LeadsTo (!vn) (F_p) (!vm))
                        )
                    )
                )
            )
        ).

(*
 * Here is the code for the main loop of the program.
 *)
Definition loop : com :=
    WHILE BNot (!T === A0) DO
        N ::= !P;
        NEW P;
        (CStore (!P) F_p (!T));
        (CStore (!P) F_n (!N));
        (CLoad Tmp_l (!T) F_l);
        (CLoad Tmp_r (!T) F_r);
        IF (BAnd (!Tmp_l === A0) (!Tmp_r === A0)) THEN
            (CIf (!I === A0)
                 (T ::= A0)
                 ( (CLoad T (!I) F_p);
                   (CLoad Tmp_l (!I) F_n);
                   DELETE !I;
                   I ::= !Tmp_l))
        ELSE
            IF (!Tmp_l === A0) THEN
                (CLoad T (!T) F_r)
            ELSE
                IF (!Tmp_r === A0) THEN
                    (CLoad T (!T) F_l)
                ELSE
                    N ::= !I;
                    NEW I;
                    (CStore (!I) F_n (!N));
                    (CLoad Tmp_l (!T) F_r);
                    (CStore (!I) F_p (!Tmp_l));
                    (CLoad T (!T) F_l)
    LOOP.

(*
 * Finally, here is the main theorem establishing the loop invariant.
 *)

Theorem loopInvariant :
    {{afterInitAssigns}}loop{{afterWhile}}.

Proof.
    (* Break up the while portion of the loop *)
    unfold loop. unfold afterWhile.

    eapply hoare_while_abs with (invariant_pi := loopPi) (invariant_spatial := loopSpatial). simpl. reflexivity.

    (* Let's separate out the sequential statements within the while *)
    eapply hoare_seq_abs. eapply hoare_seq_abs. eapply hoare_seq_abs. eapply hoare_seq_abs.
    eapply hoare_seq_abs. eapply hoare_seq_abs.

    (* Now let's break up the ifs *)
    eapply hoare_if_abs. simpl. reflexivity.
        eapply hoare_if_abs. simpl. reflexivity.

            (* Left side of the outer if, left side of the inner if T ::= A0 *)
            eapply hoare_abs_asgn. simpl. reflexivity. simpl. reflexivity.
            eapply consequent_pi_reflexive.


            (* Right side of the inner f, break up the three statements here *)
            eapply hoare_seq_abs. eapply hoare_seq_abs. eapply hoare_seq_abs.

            (* I ::= !Tmp_l *)
            eapply hoare_abs_asgn. simpl. reflexivity. simpl. reflexivity.
            eapply consequent_pi_reflexive.

            (* DELETE !I *)
            eapply hoare_abs_delete with (fmap := (F_p,(!T))::(F_n,(!Tmp_l))::nil).

            reflexivity.

            (* Tmp_l ::= !I --> F_n *)
            eapply hoare_abs_load with (nv := !(PrimedId 10)).

            simpl. reflexivity. simpl. reflexivity.
            eapply consequent_pi_reflexive.

            (* T ::= !I --> F_p *)
            eapply hoare_abs_load with (nv := !PrimedId 11).

            simpl. reflexivity. simpl. reflexivity.

            eapply consequent_pi_reflexive.
            (* Now, we need to merge the pre-conditions of the two branches of the first inner
               if *)
            apply consequent_pi_reflexive.

            (*apply consequent_pi_add_formula with (f := (FNot (!(PrimeId 11)====A0))).*)
            apply consequent_pi_pmember with (v := (PrimedId 11)) (s := (R (F_l::F_r::nil) (!RR) nil)).
            apply consequent_pi_reflexive.

            simpl.

            eapply mergeStatesFoldRight. eapply foldRule. eapply selectHeapRight. eapply selectHeapListCons.
            eapply selectHeapLeftBranch. eapply selectHeapRight. eapply selectHeapListNil. reflexivity.
            unfold mem_heap. intro e. destruct e. simpl. intros. inversion H. unfold beq_exp. destruct i. unfold beq_id. intros. inversion H. unfold beq_id. unfold beq_nat. destruct n. intros. inversion H. destruct n. intros. inversion H.
            destruct n. intros. inversion H. destruct n. intros. inversion H. destruct n. intros. inversion H. destruct n. intros. inversion H.
            destruct n. intros. inversion H. destruct n. intros. inversion H. destruct n. intros. inversion H. destruct n. intros. inversion H. destruct n. intros. inversion H.
            eapply ex_intro with (x := F_n). simpl. reflexivity. intros. inversion H. unfold beq_exp. intros. inversion H. unfold beq_exp. intros. inversion H. unfold beq_exp. intros. inversion H.
            eapply foldedHeapFieldsCons. apply foldedHeapFieldsEmpty.

            reduceFold.  eapply FoldFormulaAll. apply MemHeapsCons. apply MemHeapsNil. unfold beq_absHeap. unfold beq_id_list. unfold F_n. simpl. reflexivity.
            eapply FoldFormulaExists. apply MemHeapsCons. apply MemHeapsNil. unfold beq_absHeap. unfold beq_id_list. unfold F_n. simpl. reflexivity.
            reduceFold.  eapply FoldFormulaAll2.  apply MemHeapsFirst. auto. reflexivity.
            simpl. eapply consequent_pi_exist_consequent. simpl.
            eapply ExistsConsequent1. simpl. reflexivity. consequent_select_recurse. apply HasPmemberFirst.
            simpl. apply consequent_pi_pbody_true.
            intros. apply mem_heaps_single_list in H. subst.
            consequent_pi_select_elim.

            eapply FoldFormulaExists. apply MemHeapsCons. apply MemHeapsNil. simpl. reflexivity. reduceFold.

            eapply mergeStatesSame. diffHeap. eapply AddPMember. apply selectHeapLeftBranch. apply selectHeapRight. eapply AddPMember. apply selectHeapLeftBranch. apply selectHeapRight. eapply consequent_pi_add_formula. (*eapply consequent_pi_add_formula with (f := (!I ~ F_p ~> !PrimeId 11)). eapply consequent_pi_add_formula with (f := (!I ~ F_n ~> !PrimeId 10)).*) (*eapply consequent_pi_pmember with (v := (PrimeId 11)) (s := (R (F_l :: F_r :: nil) (!RR) nil)).*)
            (*eapply consequent_pi_add_formula with (f := (FNot (!PrimeId 11 ==== A0))).*) eapply consequent_pi_reflexive.
            simpl. eapply consequent_pi_eliminate_loaner. simpl. apply ACAnd1. apply ACAnd2. apply ACRef. simpl. reflexivity. apply HasPmemberFirst. simpl. reflexivity. reduceVarFree. simpl. apply ACAnd2. apply ACNot. simpl.
            eapply consequent_pi_eliminate_subsumed. apply HasPmemberFirst. apply DeletePmemberNext. apply DeletePmemberFirst. simpl. apply ACAnd1. apply ACRef. simpl. reduceVarFree. apply ACAnd1. apply ACAnd2. apply ACAll.
            simpl.
            eapply consequent_pi_both_pmember. eapply consequent_pi_prime_pmember. 2:reduceVarFree. eapply consequent_pi_and. simpl. reflexivity. intros. simpl. consequent_formula_and_inversion.
                 apply consequent_pi_pbody_true.
                 apply consequent_pi_pbody. consequent_formula_select_elim.
                 apply consequent_pi_pbody. consequent_formula_select_elim.
                 apply consequent_pi_pbody_true.

        (* We are now on the right branch of the outer-if, here we need to process the inner-if
           on that branch *)
        eapply hoare_if_abs.

            simpl. reflexivity.

            (* T ::= !T --> F_r *)
            eapply hoare_abs_load with (nv := !Tmp_r).
            simpl. reflexivity. simpl. reflexivity.
            apply consequent_pi_reflexive.

            eapply hoare_if_abs. simpl. reflexivity.

                (* T ::= !T --> F_l *)
                eapply hoare_abs_load with (nv := !Tmp_l).
                simpl. reflexivity. simpl. reflexivity.
                eapply consequent_pi_pmember with (v := T) (s := (R (F_l :: F_r :: nil) (!RR) nil)).
                apply consequent_pi_add_formula with (f := (!T ~ F_r ~> !Tmp_r)).
                apply consequent_pi_reflexive.

                (* Right branch of If--break up the statements *)
                eapply hoare_seq_abs. eapply hoare_seq_abs. eapply hoare_seq_abs. eapply hoare_seq_abs.
                eapply hoare_seq_abs.

                (* T ::= !T --> F_l *)
                eapply hoare_abs_load with (nv := !PrimedId 16).
                simpl. reflexivity. simpl. reflexivity.
                eapply consequent_pi_pmember with (v := T) (s := (R (F_l::F_r::nil) (!RR) nil)).
                apply consequent_pi_reflexive.

                (* We need to unfold before handling the next if *)
                eapply hoare_unfold_post. eapply unfoldRule with (var := !I). selectHeap.
                eapply buildHeapsCons with (v := N). eapply buildHeapsNil. reduceVarFree.
                reduceVarFree. reflexivity.

                unf. eapply unfAll. reflexivity. reflexivity. unfold I. unfold P. unfold beq_id. unfold beq_nat. reflexivity.
                eapply unfExists. reflexivity. reflexivity. unfold I. unfold RR. unfold beq_id. simpl. reflexivity.
                unf. eapply unfAll2. reflexivity. reflexivity. auto.
                eapply unfExists.  reflexivity. reflexivity. auto. unf.

                eapply buildAllListCons. eapply buildAllListNil. reflexivity. simpl.

                (* !I --> F_p ::= !Tmp_l *)
                eapply hoare_abs_store. selectHeap. simpl. instantiate (1 := ((F_p, _)::nil)). simpl. reflexivity.
                instantiate (1 := None). simpl. reflexivity.

                calT. eapply calTLeadsTo2. simpl. reflexivity. selectHeap.
                eapply calTLeadsTo2. simpl. reflexivity. selectHeap.

                calT. simpl.
                eapply calTLeadsTo1. simpl. reflexivity. simpl. reflexivity. reflexivity. simpl. reflexivity.

                eapply calTLeadsTo2. simpl. reflexivity. selectHeap. reflexivity.

                (* Tmp_l ::= !T --> F_r *)
                eapply hoare_abs_load with (nv := !Tmp_r).
                simpl. reflexivity. simpl. reflexivity.
                apply consequent_pi_add_formula with (f := (!T ~ F_l ~> !Tmp_l)).
                apply consequent_pi_reflexive.

                (* !I --> F_n ::= !N *)
                eapply hoare_abs_store. selectHeap. simpl. reflexivity. instantiate (1 := None). simpl. reflexivity.
                calT. eapply calTLeadsTo2. simpl. reflexivity. selectHeap.
                eapply calTLeadsTo2. simpl. reflexivity. selectHeap.
                eapply calTLeadsTo2. simpl. reflexivity. selectHeap.
                eapply calTLeadsTo2. simpl. reflexivity. selectHeap. 
                eapply calTLeadsTo2. simpl. reflexivity. selectHeap.
                reflexivity.

                (* New !I *)
                eapply hoare_abs_new.
                selectHeap.
                reduceVarFree. reduceVarFree.

                (* N ::= !I *)
                eapply hoare_abs_asgn.
                simpl. reflexivity. simpl. reflexivity.

                (* Merge the branches *)
                apply consequent_pi_reflexive. apply consequent_pi_reflexive.
                apply consequent_pi_reflexive.
                eapply mergeStatesSame. diffHeap. apply AddPMemberNone. eapply AddPMemberDropLoner.
                simpl. reflexivity. apply ACAnd1. apply ACAnd1. apply ACAnd2. apply ACRef. apply ACAnd1. apply ACRef.
                reduceVarFree. reduceVarFree. apply consequent_pi_add_formula.
                apply consequent_pi_reflexive.
                apply consequent_pi_add_formula. simpl.

                eapply consequent_pi_and. simpl. reflexivity. intros. simpl. consequent_formula_and_inversion.
                     apply consequent_pi_both_pmember. apply consequent_pi_pbody. consequent_formula_select_elim.
                     apply consequent_pi_both_pmember. apply consequent_pi_pbody. consequent_formula_select_elim.
                     apply consequent_pi_both_pmember. apply consequent_pi_pbody_true.
                     apply consequent_pi_both_pmember. apply consequent_pi_pbody. consequent_formula_select_elim.
                     apply consequent_pi_both_pmember. apply consequent_pi_pbody. consequent_formula_select_elim.
                     apply consequent_pi_exist_consequent. simpl. eapply ExistsConsequent2. 2:simpl. 2:reflexivity.
                     reduceVarFree. simpl. apply consequent_pi_both_pmember. apply consequent_pi_pbody_true.

            (* Merge of the if branches *)
            instantiate (1 := (PMember T (R (F_l::F_r::nil) (!RR) nil) (PBody
     ((((fALL vn in R (F_n :: nil) (!P) nil,
        (fExists vm in R (F_l :: F_r :: nil) (!RR) nil, (!vn ~ F_p ~> !vm)) &&&&
        fALL vn in R (F_n :: nil) (!I) nil,
        (fExists vm in R (F_l :: F_r :: nil) (!RR) nil, (!vn ~ F_p ~> !vm))) &&&&
      (!T ~ F_r ~> !Tmp_r)) &&&& (!T ~ F_l ~> !Tmp_l)))))).

      eapply consequent_pi_and. simpl. reflexivity. intros. consequent_formula_and_inversion.
          apply consequent_pi_pmember. consequent_pi_select_elim. apply consequent_pi_pmember. consequent_pi_select_elim. apply consequent_pi_pmember. consequent_pi_select_elim.
      simpl.
instantiate (1 := (PMember T (R (F_l :: F_r :: nil) (!RR) nil) (PBody
     ((((fALL vn in R (F_n :: nil) (!P) nil,
        (fExists vm in R (F_l :: F_r :: nil) (!RR) nil, (!vn ~ F_p ~> !vm)) &&&&
        fALL vn in R (F_n :: nil) (!I) nil,
        (fExists vm in R (F_l :: F_r :: nil) (!RR) nil, (!vn ~ F_p ~> !vm))) &&&&
      (!T ~ F_r ~> !Tmp_r)) &&&& (!T ~ F_l ~> !Tmp_l)) &&&& (FNot ((!Tmp_l)====A0)))))).
            eapply consequent_pi_and. simpl. reflexivity. intros. consequent_formula_and_inversion.
                apply consequent_pi_both_pmember. consequent_pi_select_elim. apply consequent_pi_both_pmember. consequent_pi_select_elim. apply consequent_pi_both_pmember. consequent_pi_select_elim.
                apply consequent_pi_both_pmember. consequent_pi_select_elim.
            eapply mergeStatesSame. diffHeap. apply AddPMemberNone. apply AddPMemberNone.

            (* Another merge of if statement branches *)
            instantiate (1 := (PMember T (R (F_l::F_r::nil) (!RR) nil) (PBody
     ((((fALL vn in R (F_n :: nil) (!P) nil,
        (fExists vm in R (F_l :: F_r :: nil) (!RR) nil, (!vn ~ F_p ~> !vm)) &&&&
        fALL vn in R (F_n :: nil) (!I) nil,
        (fExists vm in R (F_l :: F_r :: nil) (!RR) nil, (!vn ~ F_p ~> !vm))) &&&&
      (!T ~ F_r ~> !Tmp_r)) &&&& (!T ~ F_l ~> !Tmp_l)) &&&& (FOr (FNot ((!Tmp_l)====A0)) (FNot ((!Tmp_r)====A0))))))).

          simpl. apply consequent_pi_both_pmember. apply consequent_pi_pbody. apply consequent_formula_reduce_not. simpl.
          eapply consequent_formula_and. intros. 
          consequent_formula_and_inversion.
              consequent_formula_select_elim.
              consequent_formula_select_elim.
              consequent_formula_select_elim.
              consequent_formula_select_elim.
          simpl.
          eapply consequent_pi_and. reflexivity. intros. consequent_formula_and_inversion.
              apply consequent_pi_both_pmember. apply consequent_pi_pbody. consequent_formula_select_elim.
              apply consequent_pi_both_pmember. apply consequent_pi_pbody. consequent_formula_select_elim.
              apply consequent_pi_both_pmember. apply consequent_pi_pbody. consequent_formula_select_elim.
              apply consequent_pi_both_pmember. apply consequent_pi_pbody. consequent_formula_select_elim.
              apply consequent_pi_both_pmember. apply consequent_pi_pbody. consequent_formula_select_elim.

      simpl.
      instantiate (1 := (PMember T (R (F_l::F_r::nil) (!RR) nil) (PMember I (R (F_n::nil) (!I) nil) (PBody
     ((((fALL vn in R (F_n :: nil) (!P) nil,
        (fExists vm in R (F_l :: F_r :: nil) (!RR) nil, (!vn ~ F_p ~> !vm)) &&&&
        fALL vn in R (F_n :: nil) (!I) nil,
        (fExists vm in R (F_l :: F_r :: nil) (!RR) nil, (!vn ~ F_p ~> !vm))) &&&&
      (!T ~ F_r ~> !Tmp_r)) &&&& (!T ~ F_l ~> !Tmp_l))))))).
      eapply consequent_pi_and. simpl. reflexivity. intros. consequent_formula_and_inversion.
          apply consequent_pi_pmember. apply consequent_pi_both_pmember. apply consequent_pi_pbody. consequent_formula_select_elim.
          apply consequent_pi_pmember. apply consequent_pi_both_pmember. apply consequent_pi_pbody. consequent_formula_select_elim.
      apply consequent_pi_reflexive.
      eapply mergeStatesSame. diffHeap. apply AddPMemberNone. apply AddPMemberNone.
      apply consequent_pi_add_formula. apply consequent_pi_reflexive.
      simpl. apply consequent_pi_both_pmember. apply consequent_pi_pmember. apply consequent_pi_pbody. apply consequent_formula_reduce_not. simpl.
      apply consequent_formula_reflexive.

      (* Tmp_r ::= !T --> F_r *)
      eapply hoare_abs_load with (nv := (!PrimedId 12)). simpl. reflexivity. simpl. reflexivity.
      simpl. apply consequent_pi_reflexive.

      (* Tmp_l ::= !T --> F_l *)
      eapply hoare_abs_load with (nv := (!PrimedId 13)).
      simpl. reflexivity. simpl. reflexivity. apply consequent_pi_pmember with (v := T) (s := (R (F_l::F_r::nil) (!RR) nil)).
      apply consequent_pi_reflexive.

      (* unfold (R (F_n::nil) (!P) nil) *)
      eapply hoare_unfold_post.  eapply unfoldRule with (var := !P). selectHeap.

      eapply buildHeapsCons with (v := N). eapply buildHeapsNil. reduceVarFree.
      reduceVarFree. reflexivity.

      unf. eapply unfAll2. reflexivity. reflexivity. auto.
      eapply unfExists. reflexivity. reflexivity. auto. unf.

      eapply buildAllListCons. eapply buildAllListNil.

      eapply unfAll. reflexivity. reflexivity. auto.
      eapply unfExists. reflexivity. reflexivity. auto. unf. reflexivity.

      (* !P --> F_n ::= !N *)
      eapply hoare_abs_store. selectHeap. simpl. reflexivity. instantiate (1 := None). simpl. reflexivity.
      calT. eapply calTLeadsTo2. simpl. reflexivity. selectHeap.
      calT. eapply calTLeadsTo4. simpl. reflexivity. simpl. reflexivity. reflexivity.
      reflexivity. unfold F_n. unfold F_p. unfold fieldValue. simpl. reflexivity.
      eapply calTLeadsTo2. simpl. reflexivity. selectHeap.
      eapply calTLeadsTo2. simpl. reflexivity. selectHeap.
      eapply calTLeadsTo2. simpl. reflexivity. selectHeap.
      eapply calTLeadsTo2. simpl. reflexivity. selectHeap.
      eapply calTLeadsTo2. simpl. reflexivity. selectHeap.
      reflexivity.

      (* !P --> F_p ::= !T) *)
      eapply hoare_abs_store. selectHeap. instantiate (1 := ((F_p,_)::nil)). simpl. reflexivity. instantiate (1 := None). simpl. reflexivity.

      calT. eapply calTLeadsTo2. simpl. reflexivity. selectHeap.
      eapply calTLeadsTo1. simpl. reflexivity. simpl. reflexivity. reflexivity. simpl. reflexivity.
      eapply calTLeadsTo2. simpl. reflexivity. selectHeap.
      eapply calTLeadsTo2. simpl. reflexivity. selectHeap.
      eapply calTLeadsTo2. simpl. reflexivity. selectHeap.
      eapply calTLeadsTo2. simpl. reflexivity. selectHeap.
      eapply calTLeadsTo2. simpl. reflexivity. selectHeap.
      reflexivity.

      (* New P *)
      eapply hoare_abs_new. selectHeap. reduceVarFree. reduceVarFree.

      (* N ::= !P *)
      eapply hoare_abs_asgn. simpl. reflexivity. simpl. reflexivity. simpl.

      (* Completion of the hoare_abs_while conditions *)
      (*eapply consequent_pi_and. simpl. reflexivity. intros. consequent_formula_and_inversion.
          consequent_pi_select_elim. consequent_pi_select_elim.*)
      eapply consequent_pi_remove_dup_pmember. apply RemoveDupPMemberFirst. apply RemovePmember.
      apply consequent_pi_remove_redundant_pmember. simpl.
      eapply consequent_pi_eliminate_and_dup.
      simpl. apply ACAnd1. apply ACAnd2. apply ACRef. apply ACAnd2. apply ACAnd2. apply ACAnd1. apply ACAnd2. apply ACRef. simpl.
      eapply consequent_pi_eliminate_and_dup.
      apply ACAnd1. apply ACRef. apply ACAnd1. apply ACAnd2. apply ACAnd2. apply ACRef. simpl.
      eapply consequent_pi_eliminate_loaner. simpl. apply ACAnd1. apply ACAnd2. apply ACAnd2. apply ACRef. simpl. reflexivity.
      apply HasPmemberFirst. simpl. reflexivity. reduceVarFree. simpl.
      apply ACAnd2. apply ACNot. simpl.
      eapply consequent_pi_eliminate_loaner. simpl. apply ACAnd2. apply ACAnd1. apply ACAnd2. apply ACAnd2. apply ACRef.
      simpl. reflexivity. apply HasPmemberFirst. simpl. reflexivity. simpl.
      reduceVarFree. simpl. apply ACAnd2. apply ACNot. simpl.
      eapply consequent_pi_and. simpl. reflexivity. intros. simpl. consequent_formula_and_inversion.
             apply consequent_pi_both_pmember. apply consequent_pi_pbody_true.
             apply consequent_pi_both_pmember. apply consequent_pi_pbody_true.
             apply consequent_pi_both_pmember. consequent_pi_select_elim.
             apply consequent_pi_both_pmember. apply consequent_pi_exist_consequent. simpl. eapply ExistsConsequent2.
             2:simpl. 2:reflexivity. reduceVarFree. simpl. apply consequent_pi_pbody_true.
             apply consequent_pi_both_pmember. consequent_pi_select_elim.
             apply consequent_pi_both_pmember. apply consequent_pi_pbody_true.
             apply consequent_pi_both_pmember. apply consequent_pi_pbody_true.


      simpl. apply consequent_pi_both_pmember. apply consequent_pi_pbody. apply consequent_formula_reduce_not. simpl.
      eapply consequent_formula_and. intros. consequent_formula_and_inversion.
           consequent_formula_select_elim.
           consequent_formula_select_elim.
      unfold loopSpatial. apply consequent_spatial_reflexive. unfold loopPi. apply consequent_pi_reflexive.

Qed.

