(* ImpHeap.v: A simple imperative programming language and its concrete semantics *)
(* Based on work from http://www.seas.upenn.edu/~cis500/current/index.html *)

Require Export SfLib.


(* Identifiers and equivalence thereupon; used both for variables, Var/Var', 
   and field names, Fld, from the paper *)

Inductive id : Type := 
  | Id : nat -> id
  | PrimedId : nat -> id.

Definition beq_id id1 id2 :=
  match (id1, id2) with
    | (Id n1, Id n2) => beq_nat n1 n2
    | (PrimedId n1, PrimedId n2) => beq_nat n1 n2
    | _ => false
  end.

Theorem beq_id_refl : forall i,
  true = beq_id i i.
Proof.
  intros. destruct i.
  unfold beq_id. apply beq_nat_refl.  unfold beq_id. apply beq_nat_refl. Qed.

Theorem beq_id_eq : forall i1 i2,
  true = beq_id i1 i2 -> i1 = i2.
Proof. intros i1 i2. destruct i1. destruct i2. unfold beq_id. intros H. apply beq_nat_eq in H. rewrite H. reflexivity. unfold beq_id. intros H. inversion H. destruct i2. unfold beq_id. intros H. inversion H. unfold beq_id. intros H. apply beq_nat_eq in H. rewrite H.  reflexivity. Qed.


Theorem beq_id_false_not_eq : forall i1 i2,
  beq_id i1 i2 = false -> i1 <> i2.
Admitted.


Theorem not_eq_beq_id_false : forall i1 i2,
  i1 <> i2 -> beq_id i1 i2 = false.
Proof. destruct i1. destruct i2. unfold beq_id. intros H. apply not_eq_beq_false. unfold not. intros H1. rewrite H1 in H. unfold not in H. clear H1. apply H. reflexivity. intros H. unfold beq_id.  reflexivity. intros i2. destruct i2. intros H. unfold beq_id. reflexivity. unfold beq_id. intros H. apply not_eq_beq_false. unfold not. intros H1. rewrite H1 in H. unfold not in H. clear H1. apply H. reflexivity. Qed.


(* ################################################### *)

(*
 * Concrete heap and environment state definitions
 *)

(* a cell is a struct -- a partial map of field ids to heap locations *)

Definition cell := id -> option nat.

(* heap is notated h in the paper *)

Definition heap := nat -> option cell.

(* env, a mapping of id's to heap locations, is notated s in the paper *)

Definition env := id -> nat.

Definition empty_env : env := fun _ => 0.

Definition empty_heap : heap := fun _ => None.

Definition empty_cell : cell := fun _ => None.

(* the overall execution state is an environment and a heap -- h,s in the paper *)

Definition state := prod env heap.

Definition empty_state : state := (empty_env,empty_heap).

Definition env_p := fst.

Definition heap_p := snd.


(* Basic operations on States *)

Definition override (st : env) (V:id) (n : nat) :=
  fun V' => if beq_id V V' then n else st V'.

Theorem override_eq : forall n V st,
  (override st V n) V = n.
Proof.
  intros n V st.
  unfold override.
  rewrite <- beq_id_refl.
  reflexivity.
Qed.


Theorem override_neq : forall V2 V1 n st,
  beq_id V2 V1 = false ->
  (override st V2 n) V1 = (st V1).
Proof.
    intros V2 V1 n st. unfold override. intros H. rewrite H. reflexivity. Qed.

Theorem override_example : forall (n:nat), 
  (override empty_env (Id 2) n) (Id 3) = 0.
Admitted.


Theorem override_shadow : forall x1 x2 k1 k2 (f : env),
   (override  (override f k2 x1) k2 x2) k1 = (override f k2 x2) k1.
Proof. intros x1 x2 k1 k2. intros f. unfold override. remember (beq_id k2 k1). destruct b. reflexivity. reflexivity. Qed.

Theorem override_same : forall x1 k1 k2 (f : env),
  f k1 = x1 ->
  (override f k1 x1) k2 = f k2.
Admitted.

Theorem override_permute : forall x1 x2 k1 k2 k3 f,
  beq_id k2 k1 = false -> 
  (override (override f k2 x1) k1 x2) k3 = (override (override f k1 x2) k2 x1) k3.
Admitted.

(* ################################################### *)

(* The programming language syntax is now defined *)

(* Arithmetic expressions *)

Inductive aexp : Type := 
  | ANum : nat -> aexp
  | AVar : id -> aexp
  | APlus : aexp -> aexp -> aexp
  | AMinus : aexp -> aexp -> aexp
  | AMult : aexp -> aexp -> aexp.

Tactic Notation "aexp_cases" tactic(first) tactic(c) :=
  first;
  [ c "ANum" | c "AVar" | c "APlus" | c "AMinus" | c "AMult" ].

(* Shorthand for common nats and nat operators *)

Notation A0 := (ANum 0).
Notation A1 := (ANum 1).
Notation A2 := (ANum 2).
Notation A3 := (ANum 3).
Notation A4 := (ANum 4).
Notation A5 := (ANum 5).
Notation A6 := (ANum 6).

Notation "a1 '***' a2" := (AMult a1 a2)
  (at level 50).

Notation "a1 '---' a2" := (AMinus a1 a2)
  (at level 40).

Notation "a1 '+++' a2" := (APlus a1 a2)
  (at level 40).

Notation "'!' X" := (AVar X)
  (at level 30).

(* Shorthand for common variable names *)

Definition X : id := Id 0.
Definition Y : id := Id 1.
Definition Z : id := Id 2.

(* Computing equivalence of expressions *)

Fixpoint beq_exp (exp1 : aexp) (exp2 : aexp) { struct exp1 } : bool :=
  match (exp1, exp2) with
    | (AVar v1, AVar v2) => beq_id v1 v2
    | (ANum n1, ANum n2) => beq_nat n1 n2
    | (APlus a1 b1, APlus a2 b2) => if beq_exp a1 a2 then beq_exp b1 b2 else false
    | (AMinus a1 b1, APlus a2 b2) => if beq_exp a1 a2 then beq_exp b1 b2 else false
    | (AMult a1 b1, APlus a2 b2) => if beq_exp a1 a2 then beq_exp b1 b2 else false
    | _ => false
  end.

(* Boolean expressions and notation *)

Inductive bexp : Type := 
  | BTrue : bexp
  | BFalse : bexp
  | BEq : aexp -> aexp -> bexp
  | BLe : aexp -> aexp -> bexp
  | BNot : bexp -> bexp
  | BAnd : bexp -> bexp -> bexp
  | BOr : bexp -> bexp -> bexp.

Notation "b1 '===' b2" := (BEq b1 b2)
  (at level 60).

Notation "b1 '<<=' b2" := (BLe b1 b2)
  (at level 60).

Tactic Notation "bexp_cases" tactic(first) tactic(c) :=
  first;
  [ c "BTrue" | c "BFalse" | 
    c "BEq" | c "BLe" | 
    c "BNot" | c "BAnd" | c "BOr" ].

(* Commands of the language *)

Inductive com : Type :=
  | CSkip : com
  | CLoad : id -> aexp -> id -> com
  | CStore : aexp -> id -> aexp -> com
  | CAss : id -> aexp -> com
  | CNew : id -> com
  | CDelete: aexp -> com
  | CSeq : com -> com -> com
  | CIf : bexp -> com -> com -> com
  | CWhile : bexp -> com -> com.

Tactic Notation "com_cases" tactic(first) tactic(c) :=
  first;
  [ c "CSkip" | c "CLoad" | c "CStore" | c "CAss" | c "CNew" | c "CDelete" | c "CSeq" | c "CIf" | c "CWhile" ].

Notation "'SKIP'" := 
  CSkip.
Notation "c1 ; c2" := 
  (CSeq c1 c2) (at level 80, right associativity).
Notation "l '::=' v '-->' i" :=
  (CLoad l v i) (at level 60).
Notation "l '-->' i '::=' v" :=
  (CStore l i v) (at level 60).
Notation "l '::=' a" := 
  (CAss l a) (at level 60).
Notation "'NEW' v" :=
  (CNew v) (at level 60).
Notation "'DELETE' e" :=
  (CDelete e) (at level 60).
Notation "'WHILE' b 'DO' c 'LOOP'" := 
  (CWhile b c) (at level 80, right associativity).
Notation "'IF' e1 'THEN' e2 'ELSE' e3" := 
  (CIf e1 e2 e3) (at level 80, right associativity).

(* Some example programs *)

Definition plus2 : com :=
  X ::= !X +++ A2.

Example plus2_notation :
  plus2 = 
  CAss (Id 0) (APlus (AVar (Id 0)) (ANum 2)).
Proof. reflexivity. Qed.

Definition XtimesYinZ : com :=
  Z ::= !X *** !Y.

(** loops *)
Definition subtract_slowly_body : com :=
  Z ::= !Z --- A1;
  X ::= !X --- A1.

Definition subtract_slowly : com :=
  WHILE BNot (!X === A0) DO
    subtract_slowly_body
  LOOP.

Definition subtract_3_from_5_slowly : com :=
  X ::= A3;
  Z ::= A5;
  WHILE BNot (!X === A0) DO
    subtract_slowly_body
  LOOP.

(* an infinite loop *)
Definition loop : com :=
  WHILE BTrue DO
    SKIP
  LOOP.

(* factorial *)
Definition fact_body : com :=
  Y ::= !Y *** !Z;
  Z ::= !Z --- A1.

Definition fact_loop : com :=
  WHILE BNot (!Z === A0) DO
    fact_body
  LOOP.

Definition fact_com : com :=
  Z ::= !X;
  Y ::= ANum 1;
  fact_loop.

(* exponentiation *)
Definition exp_body : com :=
    Z ::= !Z *** !X;
    Y ::= !Y --- A1.

Definition pexp : com :=
  WHILE BNot (!Y === (ANum 0)) DO
    exp_body
  LOOP.

Definition add_body : com :=
  Y ::= !Y +++ !Z;
  Z ::= !Z --- A1.

Definition add_loop : com :=
  Z ::= A0;
  WHILE BNot (!Z === A0) DO
    add_body
  LOOP.

Definition is_even_body : com :=
    Z ::= A1 --- !Z.

Definition is_even : com :=
  Z ::= A0;
  WHILE BNot (!X === A0) DO
    is_even_body
  LOOP.

(* ################################################### *)

(* Now we will define what it means to evaluate a program *)

(* First, arithmetic evaluation *)

Fixpoint aeval (st : state) (e : aexp) {struct e} : nat :=
  match e with
  | ANum n => n
  | AVar i => (fst st) i
  | APlus a1 a2 => (plus (aeval st a1) (aeval st a2))
  | AMinus a1 a2 => (minus (aeval st a1) (aeval st a2))
  | AMult a1 a2 => (mult (aeval st a1) (aeval st a2))
  end.


(* Next, boolean evaluation *)

Fixpoint beval (st : state) (e : bexp) {struct e} : bool :=
  match e with 
  | BTrue       => true
  | BFalse      => false
  | BEq a1 a2   => beq_nat (aeval st a1) (aeval st a2)
  | BLe a1 a2   => ble_nat (aeval st a1) (aeval st a2)
  | BNot b1     => negb (beval st b1)
  | BAnd b1 b2  => (andb (beval st b1) (beval st b2))
  | BOr b1 b2   => (orb (beval st b1) (beval st b2))
  end.

(* Auxilliary functions needed for the evaluation relation *)

Definition bind_option (X Y : Type) (xo : option X) (f : X -> option Y) 
                      : option Y :=
  match xo with
    | None => None
    | Some x => f x
  end.

Implicit Arguments bind_option [X Y].

Fixpoint range (start : nat) (size : nat) (test : nat) { struct size } : bool :=
    match size with
      | 0 => false
      | S(s) => if beq_nat test (plus start s) then true
                else range start s test
    end.

Reserved Notation "c1 '/' st '~~>' st'" (at level 40).


Definition override_heap (h : heap) (V:nat) (c : option cell) :=
  fun V' => if beq_nat V V' then c else h V'.

Definition override_cell (c : cell) (V:id) (n : nat) :=
  fun V' => if beq_id V V' then Some n else c V'.

(* Finally, the inductive definition of evaluation of commands, giving meaning
   of concrete execution.  This is relation s, h --c--> c', h' in the paper. *)

Inductive ceval : state -> com -> state -> Prop :=
  | CESkip : forall st,
      ceval st CSkip st
  | CEAss  : forall st a1 l,
      (CAss l a1) / st ~~> ((override (fst st) l (aeval st a1)),(snd st))
  | CENew : forall st l loc,
      (not (loc=0)) ->
      ((snd st) loc)=None ->
      ((CNew l) / st ~~> (override (fst st) l loc,override_heap (snd st) loc (Some empty_cell)))
  | CEDelete : forall st loc,
      not((snd st) (aeval st loc)=None) ->
      (CDelete loc) / st ~~> (fst st,(override_heap (snd st) (aeval st loc) None))
  | CELoad : forall st loc l f fmap val,
      ((snd st) (aeval st loc)=Some fmap) ->
      (fmap f = Some val) ->
      (CLoad l loc f) / st ~~> ((override (fst st) l val),(snd st))
  | CEStore : forall st loc val fmap id,
      ((snd st) (aeval st loc)=Some fmap) ->
      (CStore loc id val) / st ~~> ((fst st),override_heap (snd st) (aeval st loc) (Some (override_cell fmap id (aeval st val))))
  | CESeq : forall c1 c2 st st' st'',
      c1 / st  ~~> st' ->
      c2 / st' ~~> st'' ->
      (CSeq c1 c2) / st ~~> st''
  | CEIfTrue : forall st st' b1 c1 c2,
      beval st b1 = true ->
      c1 / st ~~> st' ->
      (CIf b1 c1 c2) / st ~~> st'
  | CEIfFalse : forall st st' b1 c1 c2,
      beval st b1 = false ->
      c2 / st ~~> st' ->
      (CIf b1 c1 c2) / st ~~> st'
  | CEWhileEnd : forall b1 st c1,
      beval st b1 = false ->
      (CWhile b1 c1) / st ~~> st
  | CEWhileLoop : forall st st' st'' b1 c1,
      beval st b1 = true ->
      c1 / st ~~> st' ->
      (CWhile b1 c1) / st' ~~> st'' ->
      (CWhile b1 c1) / st ~~> st''

  where "c1 '/' st '~~>' st'" := (ceval st c1 st').

Tactic Notation "ceval_cases" tactic(first) tactic(c) := first; [
    c "CESkip" | c "CENew" | c "CEDelete" | c "CELoad" | c "CEStore" |
    c "CEAss" | c "CESeq" | c "CEIfTrue" | c "CEIfFalse" 
  | c "CEWhileEnd" | c "CEWhileLoop" ].



 
